
/*

   Welcome to the world of FxEngine, a 3d engine which uses 3dfx
   cards for incredible detail and smoothness.
   Key features supported (or will be supported when finished)

   - Trilinear filtered gouraud shaded texture mapping
   - Unlimited amount of RGB lightsources (moving lightsources only supported
     on objects)

   - Unlimited amount of objects
   - Unlimited size of Landscapes
   - Transparency (only in objects)
   - fogging
   - objects
   - Landscapes

   Main source file for FxEngine, I allows anybody use this code as
   much as you want, make changes and so on but you're NOT allowed to
   call the code for FxEngine without permission from the author of
   FxEngine. If you want to contact the author you can go to
   http://welcome.to/3dfxPS
   From that site you can contact the author and can get info on
   how to contribute on FxEngine and more.
   You can of course also contact him directly by e-mail:
   andreas.ingo@swipnet.se

*/

// This version is limited to 200 vertices per polygon, but you will
// probably NEVER need so much...


#include "FxEngine.h"
#include "fxlandscape.h"
#include <math.h>
#include <stdio.h>
#include <glide.h>
#include <glideutl.h>

#define MAX_VERTICES_PER_POLY 300
#define MAX_OBJECTS_IN_WORLD 10000
#define MAX_POLYGONS_PER_OBJECT 1000

fxeObjectPtr WorldObjectPointers[MAX_OBJECTS_IN_WORLD];
fxePolygonPtr PolygonPointers[MAX_POLYGONS_PER_OBJECT];

GrHwConfiguration HwConfig;

// Globals:

int CombineTextureOnly;

int ChromaKey;

float ViewingDistance;

float HorRes, VerRes; // Horisontal and Vertical Resolution

float HalfHorRes, HalfVerRes;

float HorInc, VerInc;

int HardwareSelected;


FxeDetailLevel GlobalDetailLevel;

long sqrt_table[0x100];

fxePolygon **GlobalPolygonList;

fxeObject *CurrentObject;


// All objects in world
fxeObject *WorldObjects;


int TotNumLightsources;


// World camera position and angle
fxeCamera Camera;

fxeCharactersetHandle DefaultCharacterset;

int AlphaBlendMode;

fxeColor BackgroundColor;

float SnapValue;

float CosLockup[361];
float SinLockup[361];

float NearClippingPlane;
float FarClippingPlane;



FxU32 TexMaxAddressTMU0;
FxU32 TexMaxAddressTMU1;
FxU32 TexStartAddressTMU0;
FxU32 TexStartAddressTMU1;
FxU32 TexDownloadAddressTMU0;
FxU32 TexDownloadAddressTMU1;

// All textures in world
fxeTextureHandle Textures;

int GlobalVertexIndices[3];
int GlobalVertexIndices2[3];


float HFOWInc;
float VFOWInc;

int FrameDisplayed;

// Ambient light level.
float AmbientR, AmbientG, AmbientB;

GrFog_t FxeFogTable[64];

unsigned long FrameCounter;

// Here comes some matrix math functions by Hugues Landry which also is
// going to be used in a 3d engine called Sp@cE_iNvAdr.
// Big thanks to you Hugues for this! The functions have been modified
// some by Andreas Ingo to fit better in FxEngine.

typedef float fxeMatrix[4][4];


// Here is some huge lockup tables which stores the rotation
// matrices for all rotation angles from 0 to 360.


////////////////////////////////////////////////
// Transformation Module Function declaration //
////////////////////////////////////////////////

void InitMatrix();						// identity Matrix initialization

// Matrix math Transformation Functions

void MoveMatrix(fxeMatrix Matrix, float x, float y, float z);
void RotXMatrix(fxeMatrix Matrix, float Angle);
void RotYMatrix(fxeMatrix Matrix, float Angle);
void RotZMatrix(fxeMatrix Matrix, float Angle);
void ScaleMatrix(fxeMatrix Matrix, float SFactor);
void ProjectMatrix(fxeMatrix Matrix);

extern void  CrossProduct3dASM(fxeVector *Result, fxeVector *Vector1,
  fxeVector *Vector2);
extern void SetPrecisionASM(void);


// Matrix Math multiply Functions

void Matmult(fxeMatrix MatrixA, fxeMatrix MatrixB, fxeMatrix Result);
void MatXVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result);

// Optimized Assembly version of MatXVer
void extern MatXVerASM(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result);

// This matrix is calculated for every call to fxeRenderWorld.
// fxeRenderWorld calculates the camera transformation matrix. Because
// the camera isn't changed under that function call we only need
// to calculate it once.

fxeMatrix CameraMatrix;

// Fast square root functions:
void build_sqrt_table(void) {
    unsigned short i;
    float f;
    unsigned int *fi = (unsigned*)&f;   // To access the bits of a float in
                                        // C quickly we must misuse pointers
    for (i = 0; i <= 0x7f; i++) {
        *fi = 0;

            // Build a float with the bit pattern i as mantissa
            // and an exponent of 0, stored as 127
        *fi = (i << 16) | (127 << 23);
        f = sqrt(f);

            // Take the square root then strip the first 7 bits of
            // the mantissa into the table
        sqrt_table[i] = (*fi & 0x7fffff);

            // Repeat the process, this time with an exponent of 1,
            // stored as 128
        *fi = 0;
        *fi = (i << 16) | (128 << 23);
        f = sqrt(f);
        sqrt_table[i+0x80] = (*fi & 0x7fffff);
    }
}


/* original C version */
float fast_fsqrt(float n) {
    unsigned int *num = (unsigned *)&n; // to access the bits of a float in C
                                        // we must misuse pointers

    short e;        // the exponent
    if (n == 0) return (0); /* check for square root of 0 */
    e = (*num >> 23) - 127; /* get the exponent - on a SPARC the */
                            /* exponent is stored with 127 added */
    *num &= 0x7fffff;   /* leave only the mantissa */
    if (e & 0x01) *num |= 0x800000;
                /* the exponent is odd so we have to */
                /* look it up in the second half of  */
                /* the lookup table, so we set the high bit */
    e >>= 1;    /* divide the exponent by two */
                /* note that in C the shift */
                /* operators are sign preserving */
                /* for signed operands */
        // Do the table lookup, based on the quaternary mantissa,
        // then reconstruct the result back into a float
    *num = ((sqrt_table[*num >> 16])) + ((e + 127) << 23);
    return(n);
}

void fxeFogSetMode(int FogMode)
{
  if (FogMode == NO_FOGGING)
    grFogMode(GR_FOG_DISABLE);
  else if (FogMode == DISTANCE_FOGGING)
  {
    grFogMode(GR_FOG_WITH_TABLE);
    grFogTable(FxeFogTable);
  }
}

// this function returns the number of frames that have
// been rendered since last call to fxeStartFrameCount
// if the function haven't been called the function

void fxeStartFrameCount(void)
{
  FrameCounter = 0;
}

// this function returns the number of frames that have
// been rendered since last call to fxeStartFrameCount
unsigned long fxeGetFrameCount(void)
{
  return FrameCounter;
}

void fxeFogSetColor(fxeColor *FogColor)
{
  grFogColorValue(FogColor->r * 65536+ FogColor->g * 256 + FogColor->b);
}


// This function converts a angle in degrees to a angle in radians
float DegToRad(float Deg)
{
  return Deg*0.017453292;
}

// converts a floating point value to a integer value
int FloatToInt(float Value)
{
  return (int)Value;
}

// Snaps a float value to 0.0625 resolution, required by 3dfx
// hardware
float Snap(float Value)
{
  Value += SnapValue;
  Value -= SnapValue;
  return Value;
}

// Text rendering associated functions:

void fxeCharactersetSetDefault(fxeCharactersetHandle CSH)
{
  DefaultCharacterset = CSH;
}

fxeCharactersetHandle fxeCharactersetLoad(char *FilePrefix, int Translucent,
  float AlphaLevel, int Transparent, fxeColor *TransColor, float Spacing,
  float CharWidth, float CharHeight)
{
  fxeCharactersetHandle CSH;
  int Loop;
  char Filename[256];

  CSH = (fxeCharactersetHandle)malloc(sizeof(fxeCharacterset));
  if (CSH == NULL)
    return NULL;

  for (Loop=0; Loop<256; Loop++)
  {
    sprintf(Filename,"%s%c%s",FilePrefix,(char)(Loop),".3df");

    CSH->Characterset[Loop].TextureHandle = fxeTextureLoad(Filename);

    CSH->Spacing = Spacing;
    CSH->CharWidth = CharWidth;
    CSH->CharHeight = CharHeight;

    if (CSH->Characterset[Loop].TextureHandle != NULL)
    {
      CSH->Characterset[Loop].Used = FXETRUE;
      CSH->Characterset[Loop].Translucent = Translucent;
      CSH->Characterset[Loop].AlphaLevel = AlphaLevel;
      CSH->Characterset[Loop].Transparent = Transparent;
      CSH->Characterset[Loop].TransparentColor.r = TransColor->r;
      CSH->Characterset[Loop].TransparentColor.g = TransColor->g;
      CSH->Characterset[Loop].TransparentColor.b = TransColor->b;
    }
    else
      CSH->Characterset[Loop].Used = FXEFALSE;
  }
  return CSH;
}

void fxeStringDrawXY(char *String, float X, float Y)
{
  float CurrX;

  CurrX = X;

  if (String != NULL && DefaultCharacterset != NULL)
  {
    while ((*String) != 0)
    {
      if (DefaultCharacterset->Characterset[(*String)].Used == FXETRUE)
      {
        fxeCharacterDraw(&(DefaultCharacterset->Characterset[(*String)]),
          CurrX,Y,DefaultCharacterset->CharWidth,
          DefaultCharacterset->CharHeight);
      }
      CurrX += DefaultCharacterset->Spacing;
      String++;
    }
  }
}
    
  
fxeCharacterHandle fxeCharacterCreate(fxeTextureHandle TextureHandle,
  int Translucent, float AlphaLevel, int Transparent,
  fxeColor* TransparentColor)
{
  fxeCharacter *pCharacter;


  pCharacter = (fxeCharacter*)malloc(sizeof(fxeCharacter));
  if (pCharacter == NULL)
    return NULL;

  pCharacter->TextureHandle = TextureHandle;
  pCharacter->Translucent = Translucent;
  pCharacter->AlphaLevel = AlphaLevel;
  pCharacter->Transparent = Transparent;
  pCharacter->TransparentColor.r = TransparentColor->r;
  pCharacter->TransparentColor.g = TransparentColor->g;
  pCharacter->TransparentColor.b = TransparentColor->b;
  return pCharacter;
}

void fxeCharacterDrawTranslucent(fxeCharacterHandle CharHandle,
  float X, float Y, float Width, float Height, float AlphaLevel)
{
  // Check if character is transparent:
  GrVertex Character[4];
  fxeTextureHandle TexHandle;

  TexHandle = CharHandle->TextureHandle;


  grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
    GR_BLEND_ONE,GR_BLEND_ZERO);
  AlphaBlendMode = FXETRUE;

  grConstantColorValue( (((int)AlphaLevel)<<24) );

  if (CharHandle->Transparent == FXETRUE)
  {
    grChromakeyMode(GR_CHROMAKEY_ENABLE);
    grChromakeyValue( (((int)CharHandle->TransparentColor.r) << 16) +
      (((int)CharHandle->TransparentColor.g) << 8) +
      ((int)CharHandle->TransparentColor.b) );
  }
  else if (CharHandle->Transparent == FXEFALSE)
  {
    grChromakeyMode(GR_CHROMAKEY_DISABLE);
  }

  Character[0].x = X;
  Character[0].y = VerRes-Y;
  Character[0].oow = 1.0;

  Character[1].x = X;
  Character[1].y = VerRes-(Y+Height);
  Character[1].oow = 1.0;

  Character[2].x = X+Width;
  Character[2].y = VerRes-(Y+Height);
  Character[2].oow = 1.0;

  Character[3].x = X+Width;
  Character[3].y = VerRes-Y;
  Character[3].oow = 1.0;


  grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_TEXTURE,FXFALSE);

  if (CharHandle->TextureHandle->FilteringMethod == BILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;
  }
  else if (CharHandle->TextureHandle->FilteringMethod == TRILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;

    Character[0].tmuvtx[1].sow = 0.0;
    Character[0].tmuvtx[1].tow = 0.0;
    Character[0].tmuvtx[1].oow = 1.0;
    Character[1].tmuvtx[1].sow = 0.0;
    Character[1].tmuvtx[1].tow = 256.0;
    Character[1].tmuvtx[1].oow = 1.0;
    Character[2].tmuvtx[1].sow = 256.0;
    Character[2].tmuvtx[1].tow = 256.0;
    Character[2].tmuvtx[1].oow = 1.0;
    Character[3].tmuvtx[1].sow = 256.0;
    Character[3].tmuvtx[1].tow = 0.0;
    Character[3].tmuvtx[1].oow = 1.0;
  }

  if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
  {
    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
  }

  else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
  {
    // Combine trilinear texture mapping with
    // bilinear filtering to produce trilinear
    // filtering

    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

    // Setup texture unit 1's texture source
    grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
      GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
  }

  // Draw the character:

  guDrawPolygonVertexListWithClip(4,Character);
}


void fxeCharacterDrawTransparent(fxeCharacterHandle CharHandle,
  float X, float Y, float Width, float Height, fxeColor *TransColor)
{
  // Check if character is transparent:
  GrVertex Character[4];
  fxeTextureHandle TexHandle;

  TexHandle = CharHandle->TextureHandle;

  if (CharHandle->Translucent == FXEFALSE && AlphaBlendMode == FXETRUE)
  {
    grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,
      GR_BLEND_ZERO,GR_BLEND_ZERO);
    AlphaBlendMode = FXEFALSE;
  }

  else if (CharHandle->Translucent == FXETRUE && AlphaBlendMode == FXEFALSE)
  {
    grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
      GR_BLEND_ONE,GR_BLEND_ZERO);
    AlphaBlendMode = FXETRUE;
  }

  grConstantColorValue( (((int)CharHandle->AlphaLevel)<<24) );

  grChromakeyMode(GR_CHROMAKEY_ENABLE);
  grChromakeyValue( (((int)TransColor->r) << 16) +
    (((int)TransColor->g) << 8) +
    ((int)TransColor->b) );

  Character[0].x = X;
  Character[0].y = VerRes-Y;
  Character[0].oow = 1.0;

  Character[1].x = X;
  Character[1].y = VerRes-(Y+Height);
  Character[1].oow = 1.0;

  Character[2].x = X+Width;
  Character[2].y = VerRes-(Y+Height);
  Character[2].oow = 1.0;

  Character[3].x = X+Width;
  Character[3].y = VerRes-Y;
  Character[3].oow = 1.0;


  grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_TEXTURE,FXFALSE);

  if (CharHandle->TextureHandle->FilteringMethod == BILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;
    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
  }

  else if (CharHandle->TextureHandle->FilteringMethod == TRILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;

    Character[0].tmuvtx[1].sow = 0.0;
    Character[0].tmuvtx[1].tow = 0.0;
    Character[0].tmuvtx[1].oow = 1.0;
    Character[1].tmuvtx[1].sow = 0.0;
    Character[1].tmuvtx[1].tow = 256.0;
    Character[1].tmuvtx[1].oow = 1.0;
    Character[2].tmuvtx[1].sow = 256.0;
    Character[2].tmuvtx[1].tow = 256.0;
    Character[2].tmuvtx[1].oow = 1.0;
    Character[3].tmuvtx[1].sow = 256.0;
    Character[3].tmuvtx[1].tow = 0.0;
    Character[3].tmuvtx[1].oow = 1.0;
    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

    // Setup texture unit 1's texture source
    grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
      GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
  }

  guDrawPolygonVertexListWithClip(4,Character);
}
  



void fxeCharacterDraw(fxeCharacterHandle CharHandle, float X, float Y,
  float Width, float Height)
{
  // Check if character is transparent:
  GrVertex Character[4];
  fxeTextureHandle TexHandle;

  TexHandle = CharHandle->TextureHandle;

  if (CharHandle->Translucent == FXEFALSE && AlphaBlendMode == FXETRUE)
  {
    grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,
      GR_BLEND_ZERO,GR_BLEND_ZERO);
    AlphaBlendMode = FXEFALSE;
  }

  else if (CharHandle->Translucent == FXETRUE && AlphaBlendMode == FXEFALSE)
  {
    grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
      GR_BLEND_ONE,GR_BLEND_ZERO);
    AlphaBlendMode = FXETRUE;
  }

  grConstantColorValue( (((int)CharHandle->AlphaLevel)<<24) );

  if (CharHandle->Transparent == FXETRUE && ChromaKey == FXEFALSE)
  {
    grChromakeyMode(GR_CHROMAKEY_ENABLE);
    grChromakeyValue( (((int)CharHandle->TransparentColor.r) << 16) +
      (((int)CharHandle->TransparentColor.g) << 8) +
      ((int)CharHandle->TransparentColor.b) );
    ChromaKey == FXETRUE;
  }
  else if (CharHandle->Transparent == FXEFALSE && ChromaKey == FXETRUE)
  {
    grChromakeyMode(GR_CHROMAKEY_DISABLE);
    ChromaKey == FXEFALSE;
  }

  Character[0].x = X*HorInc;
  Character[0].y = VerRes-Y*VerInc;
  Character[0].oow = 1.0;

  Character[1].x = X*HorInc;
  Character[1].y = VerRes-(Y+Height)*VerInc;
  Character[1].oow = 1.0;

  Character[2].x = (X+Width)*HorInc;
  Character[2].y = VerRes-(Y+Height)*VerInc;
  Character[2].oow = 1.0;

  Character[3].x = (X+Width)*HorInc;
  Character[3].y = VerRes-Y*VerInc;
  Character[3].oow = 1.0;

  if (CombineTextureOnly == FXEFALSE)
  {
    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,GR_COMBINE_FACTOR_ONE,
      GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_TEXTURE,FXFALSE);
    CombineTextureOnly = FXETRUE;
  }


  if (CharHandle->TextureHandle->FilteringMethod == BILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;
  }
  else if (CharHandle->TextureHandle->FilteringMethod == TRILINEAR_FILTERED)
  {
    // Copy texture coordinates:

    Character[0].tmuvtx[0].sow = 0.0;
    Character[0].tmuvtx[0].tow = 0.0;
    Character[0].tmuvtx[0].oow = 1.0;
    Character[1].tmuvtx[0].sow = 0.0;
    Character[1].tmuvtx[0].tow = 256.0;
    Character[1].tmuvtx[0].oow = 1.0;
    Character[2].tmuvtx[0].sow = 256.0;
    Character[2].tmuvtx[0].tow = 256.0;
    Character[2].tmuvtx[0].oow = 1.0;
    Character[3].tmuvtx[0].sow = 256.0;
    Character[3].tmuvtx[0].tow = 0.0;
    Character[3].tmuvtx[0].oow = 1.0;

    Character[0].tmuvtx[1].sow = 0.0;
    Character[0].tmuvtx[1].tow = 0.0;
    Character[0].tmuvtx[1].oow = 1.0;
    Character[1].tmuvtx[1].sow = 0.0;
    Character[1].tmuvtx[1].tow = 256.0;
    Character[1].tmuvtx[1].oow = 1.0;
    Character[2].tmuvtx[1].sow = 256.0;
    Character[2].tmuvtx[1].tow = 256.0;
    Character[2].tmuvtx[1].oow = 1.0;
    Character[3].tmuvtx[1].sow = 256.0;
    Character[3].tmuvtx[1].tow = 0.0;
    Character[3].tmuvtx[1].oow = 1.0;
  }

  if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
  {
    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
  }

  else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
  {
    // Combine trilinear texture mapping with
    // bilinear filtering to produce trilinear
    // filtering

    // Setup texture unit 0's texture source
    grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
      GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

    // Setup texture unit 1's texture source
    grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
      GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
  }

  // Draw the character:

  guDrawPolygonVertexListWithClip(4,Character);

}






// Landscape rendering functions:


void ClipLandscapeToNearPlane(fxeLandscapePtr pLandscape)
{
  fxeColor *ClippedVertexColors1;
  fxeColor *ClippedVertexColors2;
  fxeColor *VertexColors;
  fxeColor *CurrVertexColor;
  int Loop;
  fxeVertex *VerticesCamera;
  fxeVertex *CurrVertex;
  fxeTextureCoordinate *TextureCoordinates;
  fxeTextureCoordinate *CurrTexCoor;
  float MultValue;
  int NumVertices;
  int DestVertex;
  int Nextvert;
  int Curin, Nextin;
  int CurZ, NextZ;
  fxeVertex *ClippedVertices1;
  fxeTextureCoordinate *ClippedTextureCoordinates1;
  fxeVertex *ClippedVertices2;
  fxeTextureCoordinate *ClippedTextureCoordinates2;

  fxeVertex ExtraVertices[3];
  fxeColor ExtraVertexColors[3];
  fxeTextureCoordinate ExtraTextureCoordinates[3];

  // We needs to split the polygon in two halvs since this
  // routine only can clip convex planar polygons.


  VerticesCamera = pLandscape->VerticesCamera;
  VertexColors = pLandscape->VertexColors;
  TextureCoordinates = pLandscape->TextureCoordinates;
  ClippedVertices1 = pLandscape->ClippedVertices1;
  ClippedTextureCoordinates1 = pLandscape->ClippedTextureCoordinates1;
  ClippedVertexColors1 = pLandscape->ClippedVertexColors1;
  ClippedVertices2 = pLandscape->ClippedVertices2;
  ClippedTextureCoordinates2 = pLandscape->ClippedTextureCoordinates2;
  ClippedVertexColors2 = pLandscape->ClippedVertexColors2;

  ExtraVertices[0].x = VerticesCamera[0].x;
  ExtraVertices[0].y = VerticesCamera[0].y;
  ExtraVertices[0].z = VerticesCamera[0].z;

  ExtraVertices[1].x = VerticesCamera[2].x;
  ExtraVertices[1].y = VerticesCamera[2].y;
  ExtraVertices[1].z = VerticesCamera[2].z;

  ExtraVertices[2].x = VerticesCamera[3].x;
  ExtraVertices[2].y = VerticesCamera[3].y;
  ExtraVertices[2].z = VerticesCamera[3].z;

  ExtraVertexColors[0].r = VertexColors[0].r;
  ExtraVertexColors[0].g = VertexColors[0].g;
  ExtraVertexColors[0].b = VertexColors[0].b;

  ExtraVertexColors[1].r = VertexColors[2].r;
  ExtraVertexColors[1].g = VertexColors[2].g;
  ExtraVertexColors[1].b = VertexColors[2].b;

  ExtraVertexColors[2].r = VertexColors[3].r;
  ExtraVertexColors[2].g = VertexColors[3].g;
  ExtraVertexColors[2].b = VertexColors[3].b;

  ExtraTextureCoordinates[0].s = TextureCoordinates[0].s;
  ExtraTextureCoordinates[0].t = TextureCoordinates[0].t;

  ExtraTextureCoordinates[1].s = TextureCoordinates[2].s;
  ExtraTextureCoordinates[1].t = TextureCoordinates[2].t;

  ExtraTextureCoordinates[2].s = TextureCoordinates[3].s;
  ExtraTextureCoordinates[2].t = TextureCoordinates[3].t;

  // Go through all edges of the polygon

  pLandscape->NumAddedVertices1 = 0;
  DestVertex = 0;
  NumVertices = 3;
  CurZ = VerticesCamera[0].z;
  Curin = (CurZ >= NearClippingPlane);

  for (Loop=0; Loop<3; Loop++)
  {
    Nextvert = (Loop + 1) % NumVertices;

    if (Curin)
    {
      CurrVertex = &(VerticesCamera[Loop]);
      CurrVertexColor = &(VertexColors[Loop]);

      ClippedVertices1[DestVertex].x = CurrVertex->x;
      ClippedVertices1[DestVertex].y = CurrVertex->y;
      ClippedVertices1[DestVertex].z = CurrVertex->z;
            
      if (pLandscape->TextureMode != NOT_TEXTURED)
      {
        CurrTexCoor = &(TextureCoordinates[Loop]);
        ClippedTextureCoordinates1[DestVertex].s = CurrTexCoor->s;
        ClippedTextureCoordinates1[DestVertex].t = CurrTexCoor->t;
      }

      ClippedVertexColors1[DestVertex].r = CurrVertexColor->r;
      ClippedVertexColors1[DestVertex].g = CurrVertexColor->g;
      ClippedVertexColors1[DestVertex].b = CurrVertexColor->b;

      pLandscape->NumAddedVertices1++;

      DestVertex++;
    }

    NextZ = VerticesCamera[Nextvert].z;
    Nextin = (NextZ >= NearClippingPlane);

    if (Curin != Nextin)
    {
      MultValue = (NearClippingPlane - CurZ) / (NextZ - CurZ);

      CurrVertex = &(VerticesCamera[Loop]);
      CurrVertexColor = &(VertexColors[Loop]);

      ClippedVertices1[DestVertex].x = CurrVertex->x +
        ((VerticesCamera[Nextvert].x - CurrVertex->x)
        * MultValue);

      ClippedVertices1[DestVertex].y = CurrVertex->y +
        ((VerticesCamera[Nextvert].y - CurrVertex->y)
        * MultValue);
            
      ClippedVertices1[DestVertex].z = CurrVertex->z +
        ((VerticesCamera[Nextvert].z - CurrVertex->z)
        * MultValue);

      if (pLandscape->TextureMode != NOT_TEXTURED)
      {
        CurrTexCoor = &(TextureCoordinates[Loop]);

        ClippedTextureCoordinates1[DestVertex].s = CurrTexCoor->s +
          ((TextureCoordinates[Nextvert].s - CurrTexCoor->s)
          * MultValue);

        ClippedTextureCoordinates1[DestVertex].t = CurrTexCoor->t +
          ((TextureCoordinates[Nextvert].t - CurrTexCoor->t)
          * MultValue);
      }

      ClippedVertexColors1[DestVertex].r = CurrVertexColor->r;
      ClippedVertexColors1[DestVertex].g = CurrVertexColor->g;
      ClippedVertexColors1[DestVertex].b = CurrVertexColor->b;

      pLandscape->NumAddedVertices1++;

      DestVertex++;
    }

    CurZ = NextZ;
    Curin = Nextin;
  }

  pLandscape->NumAddedVertices2 = 0;
  DestVertex = 0;
  NumVertices = 3;

  CurZ = ExtraVertices[0].z;
  Curin = (CurZ >= NearClippingPlane);

  for (Loop=0; Loop<3; Loop++)
  {
    Nextvert = (Loop + 1) % NumVertices;

    if (Curin)
    {
      CurrVertex = &(ExtraVertices[Loop]);
      CurrVertexColor = &(ExtraVertexColors[Loop]);

      ClippedVertices2[DestVertex].x = CurrVertex->x;
      ClippedVertices2[DestVertex].y = CurrVertex->y;
      ClippedVertices2[DestVertex].z = CurrVertex->z;
            
      if (pLandscape->TextureMode != NOT_TEXTURED)
      {
        CurrTexCoor = &(ExtraTextureCoordinates[Loop]);
        ClippedTextureCoordinates2[DestVertex].s = CurrTexCoor->s;
        ClippedTextureCoordinates2[DestVertex].t = CurrTexCoor->t;
      }

      ClippedVertexColors2[DestVertex].r = CurrVertexColor->r;
      ClippedVertexColors2[DestVertex].g = CurrVertexColor->g;
      ClippedVertexColors2[DestVertex].b = CurrVertexColor->b;

      pLandscape->NumAddedVertices2++;

      DestVertex++;
    }

    NextZ = ExtraVertices[Nextvert].z;
    Nextin = (NextZ >= NearClippingPlane);

    if (Curin != Nextin)
    {
      MultValue = (NearClippingPlane - CurZ) / (NextZ - CurZ);

      CurrVertex = &(ExtraVertices[Loop]);
      CurrVertexColor = &(ExtraVertexColors[Loop]);

      ClippedVertices2[DestVertex].x = CurrVertex->x +
        ((ExtraVertices[Nextvert].x - CurrVertex->x)
        * MultValue);

      ClippedVertices2[DestVertex].y = CurrVertex->y +
        ((ExtraVertices[Nextvert].y - CurrVertex->y)
        * MultValue);
            
      ClippedVertices2[DestVertex].z = CurrVertex->z +
        ((ExtraVertices[Nextvert].z - CurrVertex->z)
        * MultValue);

      if (pLandscape->TextureMode != NOT_TEXTURED)
      {
        CurrTexCoor = &(ExtraTextureCoordinates[Loop]);

        ClippedTextureCoordinates2[DestVertex].s = CurrTexCoor->s +
          ((ExtraTextureCoordinates[Nextvert].s - CurrTexCoor->s)
          * MultValue);

        ClippedTextureCoordinates2[DestVertex].t = CurrTexCoor->t +
          ((ExtraTextureCoordinates[Nextvert].t - CurrTexCoor->t)
          * MultValue);
      }

      ClippedVertexColors2[DestVertex].r = CurrVertexColor->r;
      ClippedVertexColors2[DestVertex].g = CurrVertexColor->g;
      ClippedVertexColors2[DestVertex].b = CurrVertexColor->b;

      pLandscape->NumAddedVertices2++;

      DestVertex++;
    }

    CurZ = NextZ;
    Curin = Nextin;
  }

}

void ClipLandscape(fxeLandscapePtr pLandscape)
{
  int Loop;
  fxeLandscapePtr pLandscapeLocal = pLandscape;
  int ClippedZ1, ClippedZ2, ClippedX1, ClippedX2, ClippedY1, ClippedY2;

  ClippedZ1 = 0;
  ClippedZ2 = 0;
  ClippedX1 = 0;
  ClippedX2 = 0;
  ClippedY1 = 0;
  ClippedY2 = 0;

  pLandscapeLocal->Clipped = FXEFALSE;
  pLandscapeLocal->ClippedNearPlane = FXEFALSE;

  // Transform all vertices to camera coordinates and clip those vertices:
  for (Loop=0; Loop<4; Loop++)
  {
    MatXVerASM(CameraMatrix,&(pLandscapeLocal->VerticesWorld[Loop]),
      &(pLandscapeLocal->VerticesCamera[Loop]));

    if (pLandscapeLocal->VerticesCamera[Loop].z < NearClippingPlane)
      ClippedZ1++;
    if (pLandscapeLocal->VerticesCamera[Loop].z > FarClippingPlane)
      ClippedZ2++;
    if (pLandscapeLocal->VerticesCamera[Loop].x < (-(pLandscapeLocal->
      VerticesCamera[Loop].z)))
      ClippedX1++;
    if (pLandscapeLocal->VerticesCamera[Loop].x > (pLandscapeLocal->
      VerticesCamera[Loop].z))
      ClippedX2++;
    if (pLandscapeLocal->VerticesCamera[Loop].y <
      (-(pLandscapeLocal->VerticesCamera[Loop].z * VFOWInc)))
      ClippedY1++;
    if (pLandscapeLocal->VerticesCamera[Loop].y > (pLandscapeLocal->
      VerticesCamera[Loop].z * VFOWInc))
      ClippedY2++;
  }

  if (ClippedZ1 > 0 || ClippedZ2 > 0 || ClippedX1 > 0 || ClippedX2 > 0
    || ClippedY1 > 0 || ClippedY2 > 0)
    pLandscapeLocal->Clipped = FXETRUE;

  if (ClippedZ1 > 0)
  {
    pLandscapeLocal->ClippedNearPlane = FXETRUE;
    ClipLandscapeToNearPlane(pLandscapeLocal);
  }
}

void ProjectAndDrawLandscape(fxeLandscapePtr pLandscape)
{
  int Loop;
  GrVertex ProjectedVertices[8];
  float oow;
  float oowView;
  fxeLandscapePtr pLandscapeLocal = pLandscape;
  float px, py;
  fxeTextureHandle TexHandle;

  if (pLandscapeLocal->ClippedNearPlane == FXETRUE)
  {

    for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices1; Loop++)
    {
      // Transform camera vertices to projected vertices:
      oow = 1.0 / pLandscapeLocal->ClippedVertices1[Loop].z;
      oowView = oow * ViewingDistance;

      px = pLandscapeLocal->ClippedVertices1[Loop].x *
        oowView + HalfHorRes;

      py = pLandscapeLocal->ClippedVertices1[Loop].y *
        oowView + HalfVerRes;


      px += SnapValue;
      px -= SnapValue;
      py += SnapValue;
      py -= SnapValue;

      ProjectedVertices[Loop].x = px;
      ProjectedVertices[Loop].y = py;
      ProjectedVertices[Loop].oow = oow;

      // Transfer vertex colors:

      ProjectedVertices[Loop].r =
        pLandscapeLocal->ClippedVertexColors1[Loop].r + AmbientR;


      if (ProjectedVertices[Loop].r > 255.0)
        ProjectedVertices[Loop].r = 255.0;


      ProjectedVertices[Loop].g =
        pLandscapeLocal->ClippedVertexColors1[Loop].g + AmbientG;

      if (ProjectedVertices[Loop].g > 255.0)
        ProjectedVertices[Loop].g = 255.0;

      ProjectedVertices[Loop].b =
        pLandscapeLocal->ClippedVertexColors1[Loop].b + AmbientB;

      if (ProjectedVertices[Loop].b > 255.0)
        ProjectedVertices[Loop].b = 255.0;

      // Transfer texture coordinates if needed
      if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
      {
        if (pLandscapeLocal->TH->FilteringMethod == BILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].t * oow;
        }
        else if (pLandscapeLocal->TH->FilteringMethod == TRILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].t * oow;
          ProjectedVertices[Loop].tmuvtx[1].sow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[1].tow =
            pLandscapeLocal->ClippedTextureCoordinates1[Loop].t * oow;
        }
      }
    }

    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
      {
         // Setup texture unit 0's texture source
         grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
           GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
      }

      else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
      {

        // Setup texture unit 0's texture source
        grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
          GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

        // Setup texture unit 1's texture source
        grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
          GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
      }
    }

    if (pLandscapeLocal->NumAddedVertices1==3)
    {
      guDrawTriangleWithClip(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
    }
    else if (pLandscapeLocal->NumAddedVertices1==4)
    {
      guDrawTriangleWithClip(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
      guDrawTriangleWithClip(&(ProjectedVertices[2]),&(ProjectedVertices[3]),
        &(ProjectedVertices[0]));
    }



    for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices2; Loop++)
    {
      // Transform camera vertices to projected vertices:
      oow = 1.0 / pLandscapeLocal->ClippedVertices2[Loop].z;
      oowView = oow * ViewingDistance;

      px = pLandscapeLocal->ClippedVertices2[Loop].x *
        oowView + HalfHorRes;

      py = pLandscapeLocal->ClippedVertices2[Loop].y *
        oowView + HalfVerRes;


      px += SnapValue;
      px -= SnapValue;
      py += SnapValue;
      py -= SnapValue;

      ProjectedVertices[Loop].x = px;
      ProjectedVertices[Loop].y = py;
      ProjectedVertices[Loop].oow = oow;

      // Transfer vertex colors:

      ProjectedVertices[Loop].r =
        pLandscapeLocal->ClippedVertexColors2[Loop].r + AmbientR;


      if (ProjectedVertices[Loop].r > 255.0)
        ProjectedVertices[Loop].r = 255.0;


      ProjectedVertices[Loop].g =
        pLandscapeLocal->ClippedVertexColors2[Loop].g + AmbientG;

      if (ProjectedVertices[Loop].g > 255.0)
        ProjectedVertices[Loop].g = 255.0;

      ProjectedVertices[Loop].b =
        pLandscapeLocal->ClippedVertexColors2[Loop].b + AmbientB;

      if (ProjectedVertices[Loop].b > 255.0)
        ProjectedVertices[Loop].b = 255.0;

      // Transfer texture coordinates if needed
      if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
      {
        if (pLandscapeLocal->TH->FilteringMethod == BILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].t * oow;
        }
        else if (pLandscapeLocal->TH->FilteringMethod == TRILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].t * oow;
          ProjectedVertices[Loop].tmuvtx[1].sow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[1].tow =
            pLandscapeLocal->ClippedTextureCoordinates2[Loop].t * oow;
        }
      }
    }

    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
      {
         // Setup texture unit 0's texture source
         grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
           GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
      }

      else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
      {

        // Setup texture unit 0's texture source
        grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
          GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

        // Setup texture unit 1's texture source
        grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
          GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
      }
    }

    if (pLandscapeLocal->NumAddedVertices2==3)
    {
      guDrawTriangleWithClip(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
    }
    else if (pLandscapeLocal->NumAddedVertices2==4)
    {
      guDrawTriangleWithClip(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
      guDrawTriangleWithClip(&(ProjectedVertices[2]),&(ProjectedVertices[3]),
        &(ProjectedVertices[0]));
    }


  }

  if (pLandscapeLocal->ClippedNearPlane == FXEFALSE)
  {

    for (Loop=0; Loop<4; Loop++)
    {
      // Transform camera vertices to projected vertices:

      oow = 1.0 / pLandscapeLocal->VerticesCamera[Loop].z;
      oowView = oow * ViewingDistance;

      px = pLandscapeLocal->VerticesCamera[Loop].x *
        oowView + HalfHorRes;

      py = pLandscapeLocal->VerticesCamera[Loop].y *
        oowView + HalfVerRes;


      px += SnapValue;
      px -= SnapValue;
      py += SnapValue;
      py -= SnapValue;

      ProjectedVertices[Loop].x = px;
      ProjectedVertices[Loop].y = py;
      ProjectedVertices[Loop].oow = oow;

      // Transfer vertex colors and add ambient light to the values,
      // vertex color values is clipped to 255.0

      ProjectedVertices[Loop].r =
        pLandscapeLocal->VertexColors[Loop].r + AmbientR;

      if (ProjectedVertices[Loop].r > 255.0)
        ProjectedVertices[Loop].r = 255.0;


      ProjectedVertices[Loop].g =
        pLandscapeLocal->VertexColors[Loop].g + AmbientG;

      if (ProjectedVertices[Loop].g > 255.0)
        ProjectedVertices[Loop].g = 255.0;

      ProjectedVertices[Loop].b =
        pLandscapeLocal->VertexColors[Loop].b + AmbientB;

      if (ProjectedVertices[Loop].b > 255.0)
        ProjectedVertices[Loop].b = 255.0;

      // Transfer texture coordinates if needed

      if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
      {
        if (pLandscapeLocal->TH->FilteringMethod == BILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->TextureCoordinates[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->TextureCoordinates[Loop].t * oow;
        }
        else if (pLandscapeLocal->TH->FilteringMethod == TRILINEAR_FILTERED)
        {
          ProjectedVertices[Loop].tmuvtx[0].sow =
            pLandscapeLocal->TextureCoordinates[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[0].tow =
            pLandscapeLocal->TextureCoordinates[Loop].t * oow;
          ProjectedVertices[Loop].tmuvtx[1].sow =
            pLandscapeLocal->TextureCoordinates[Loop].s * oow;
          ProjectedVertices[Loop].tmuvtx[1].tow =
            pLandscapeLocal->TextureCoordinates[Loop].t * oow;
        }
      }
    }


    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
      {
         // Setup texture unit 0's texture source
         grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
           GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
      }

      else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
      {

        // Setup texture unit 0's texture source
        grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
          GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

        // Setup texture unit 1's texture source
        grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
          GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
      }
    }

    // Draw triangles:

    if (pLandscapeLocal->Clipped == FXETRUE)
    {
      guDrawTriangleWithClip(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
      guDrawTriangleWithClip(&(ProjectedVertices[2]),&(ProjectedVertices[3]),
        &(ProjectedVertices[0]));
    }
    else
    {
      grDrawTriangle(&(ProjectedVertices[0]),&(ProjectedVertices[1]),
        &(ProjectedVertices[2]));
      grDrawTriangle(&(ProjectedVertices[2]),&(ProjectedVertices[3]),
        &(ProjectedVertices[0]));
    }
  }
}

void RenderLandscape(fxeLandscapePtr pLandscape)
{
  int Loop,Loop2;
  fxeVertex *VertexCamera;
  float Radius;

  // Transform world pos of landscape to camera coordinates
  VertexCamera = &(pLandscape->VertexCamera);
  Radius = pLandscape->Radius;


  MatXVerASM(CameraMatrix,&(pLandscape->VertexWorld),VertexCamera);

  if (VertexCamera->z + Radius < NearClippingPlane ||
      VertexCamera->z - Radius > FarClippingPlane)
    return;
  else if (VertexCamera->x + Radius < (-VertexCamera->z) ||
           VertexCamera->x - Radius > VertexCamera->z)
    return;
  else if (VertexCamera->y + Radius < (-VertexCamera->z*VFOWInc) ||
           VertexCamera->y - Radius > (VertexCamera->z*VFOWInc))
    return;


  if (pLandscape->Landscapes != NULL)
  {
    for (Loop=0 ; Loop<pLandscape->NumLandscapesX; Loop++)
    {
      for (Loop2=0 ; Loop2<pLandscape->NumLandscapesX; Loop2++)
      {
        RenderLandscape((fxeLandscapePtr)
          (pLandscape->Landscapes
          [(Loop*(pLandscape->NumLandscapesX)+Loop2)]));
      }
    }
  }
  else
  {
    // Draw Polygon directly since this Landscape doesn't have child
    // landscapes:

    ClipLandscape(pLandscape);
    ProjectAndDrawLandscape(pLandscape);
  }
}


  

void fxeLandscapeRender(fxeLandscapeHandle pLandscape)
{
  fxeLandscapePtr pLandscapeLocal = pLandscape;


  grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,
    GR_BLEND_ZERO,GR_BLEND_ZERO);
  AlphaBlendMode = FXEFALSE;
  
  // Setup color combine unit for gouraud shading
  // and texture mapping:

  if (pLandscape->TextureMode == TEXTURE_MAPPED)
  {
    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
      GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
      GR_COMBINE_OTHER_TEXTURE,FXFALSE);
  }
  else if (pLandscape->TextureMode == NOT_TEXTURED)
  {
    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,GR_COMBINE_LOCAL_ITERATED,
      GR_COMBINE_OTHER_NONE,FXFALSE);
  }

  RenderLandscape(pLandscapeLocal);
}

  
void fxeGetVersion(int *MainVersion, int *UnderVersion)
{
  *MainVersion = 0;
  *UnderVersion = 3;
}

// call this function first - initializes FxEngine
// Returns FXETRUE if successful and FXEFALSE if failure
int fxeInitialize(FxeDetailLevel DetailLevel)
{
  int sliDetect;
  GrScreenResolution_t GlideRes;
  FxU32 FCW;
  GrVoodooConfig_t *Config;
  int Loop;


  // Initialize glide:
  grGlideInit();

  // Check for 3dfx hardware:
  if (grSstQueryHardware(&HwConfig) == FXFALSE)
  {
    return FXEFALSE;
  }


  // Select first 3dfx hardware found.
  HardwareSelected = 0;
  grSstSelect(HardwareSelected);

  // check which hardware (voodoo and voodoo2 supported)
  // I now this is a bad way to do this but since glide didn't report
  // my 3dblaster voodoo2 as a voodoo2 then this way is the only
  // way to go.
  if (HwConfig.SSTs[HardwareSelected].type == GR_SSTTYPE_VOODOO)
    Config = &(HwConfig.SSTs[HardwareSelected].sstBoard.VoodooConfig);
  if (HwConfig.SSTs[HardwareSelected].type == GR_SSTTYPE_Voodoo2)
    Config = &(HwConfig.SSTs[HardwareSelected].sstBoard.Voodoo2Config);


  if (Config->nTexelfx == 1)
  {
    // We now it's a voodoo because it has only 
    // one TexelFx.
    // when the FxEngine is run on a voodoo card the detail
    // settings must be the lowest possible


    if (DetailLevel != LOW_DETAIL && DetailLevel != AUTOMATIC_DETAIL)
    {
      grGlideShutdown();
      return FXEFALSE;
    }

    GlobalDetailLevel = LOW_DETAIL;
    HorRes = 640.0;
    VerRes = 480.0;
    GlideRes = GR_RESOLUTION_640x480;
  }

  else if (Config->nTexelfx == 2 && Config->fbRam >= 4)
  {
    // We know it must be a voodoo2 because it has two TexelFx units and
    // 4mb or more frame buffer memory
  sliDetect = Config->sliDetect;
    // Check resolution, if 1024x768 we must check for SLI
	if (sliDetect == FXFALSE && DetailLevel == MAXIMUM_DETAIL)
	{
	  grGlideShutdown();
	  return FXEFALSE;
    }
	if (DetailLevel == AUTOMATIC_DETAIL && sliDetect == FXTRUE)
	{
	  GlobalDetailLevel = MAXIMUM_DETAIL;
	  HorRes = 1024;
	  VerRes = 768;
	  GlideRes = GR_RESOLUTION_1024x768;
  }
	else if (DetailLevel == AUTOMATIC_DETAIL && sliDetect == FXFALSE)
  {
	  GlobalDetailLevel = HIGH_DETAIL;
    HorRes = 800;
    VerRes = 600;
	  GlideRes = GR_RESOLUTION_800x600;
	}
	else
	{
	  GlobalDetailLevel = DetailLevel;
	  if (DetailLevel == LOW_DETAIL)
	  {
      HorRes = 640.0;
      VerRes = 480.0;
	  	GlideRes = GR_RESOLUTION_640x480;
      }
	  else if (DetailLevel = HIGH_DETAIL)
	  {
	    HorRes = 800;
      VerRes = 600;
	    GlideRes = GR_RESOLUTION_800x600;
	  }
	  else
	  {
	    HorRes = 1024;
      VerRes = 768;
	  	GlideRes = GR_RESOLUTION_1024x768;
	  }
    }
  }
  else
  {

    // When the SstType is unknown the detaillevel and resolution
    // defaults to lowest possible

    if (DetailLevel != LOW_DETAIL && DetailLevel != AUTOMATIC_DETAIL)
    {
      grGlideShutdown();
      return FXEFALSE;
    }


    GlobalDetailLevel = LOW_DETAIL;
    HorRes = 640.0;
    VerRes = 480.0;
    GlideRes = GR_RESOLUTION_640x480;
  }

  // After the resolution and detaillevel varibles have been
  // we swithes to 3dfx videomode using a refresh rate of 60HZ,
  // ARGB colorformat, double buffering and a auxiliary buffer
  // which is going to be used for z buffering

  if (grSstWinOpen(0,GlideRes,GR_REFRESH_60Hz,GR_COLORFORMAT_ARGB,
    GR_ORIGIN_LOWER_LEFT,2,1) == FXFALSE)
  {
    grGlideShutdown();
    return FXEFALSE;
  }

  if (GlobalDetailLevel == LOW_DETAIL)
  {
    // Set up texture unit 0 for bilinear filtering
    grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_BILINEAR,
      GR_TEXTUREFILTER_BILINEAR);

    // Setup TMU0 for simple texture mapping
    grTexCombine(GR_TMU0,GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      FXFALSE,FXFALSE);
  }

  else
  {
    // Set up texture unit 0 for bilinear filtering
    grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_BILINEAR,
      GR_TEXTUREFILTER_BILINEAR);

    // Set up texture unit 1 for bilinear filtering
    grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_BILINEAR,
      GR_TEXTUREFILTER_BILINEAR);

    // Setup TMU1 for simple texture mapping - the
    // data from TMU1 is blended with texture unit
    // 0's data to produce trilinear texture mapping
    grTexCombine(GR_TMU1,GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      FXFALSE,FXFALSE);

    grTexCombine(GR_TMU0,
      GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
      GR_COMBINE_FACTOR_LOD_FRACTION,
      GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
      GR_COMBINE_FACTOR_LOD_FRACTION,
      FXFALSE,FXFALSE);
  }
    

  // Now Inititialise some structures and variabled used by
  // FxEngine

  // Setup the fogtable:

  for (Loop=0 ; Loop<64 ; Loop++)
  {
    FxeFogTable[Loop] = 255;
  }

  // Build sqrt table:
  build_sqrt_table();

  for (Loop=0 ; Loop<30; Loop++)
  {
    FxeFogTable[Loop] = 0;
  }

  for (Loop=30 ; Loop<45 ; Loop++)
  {
    FxeFogTable[Loop] = (GrFog_t)( (float)(Loop-30)*(255.0/15) );
  }
  // Reset camera position to 0,0,0 with angles 0,0,0
  Camera.x = 0;
  Camera.y = 0;
  Camera.z = 0;
  Camera.ax = 0;
  Camera.ay = 0;
  Camera.az = 0;

  // Reset world objects:

  WorldObjects = NULL;

  // Reset Lightsources:

  Lightsources = NULL;

  // Reset Textures:
  Textures = NULL;


  // Calculate viewing distance such as the horisontal field of view
  // is 90 degrees independent of resolution

  ViewingDistance = (HorRes/2.0) / tan(DegToRad(45.0));

  HFOWInc = (HorRes/2.0) / ViewingDistance;
  VFOWInc = (VerRes/2.0) / ViewingDistance;


  // Configure floating point unit in microprocessor to not change
  // precision, this is required to make the Snap() routine to work
  SnapValue = (float) (3L << 18);

  // Set FCW to 0 to avoid warnings
  FCW = 0;

  SetPrecisionASM();


  // Precalculate cos and sin tables:
  for (Loop=0; Loop<361; Loop++)
  {
    CosLockup[Loop] = cos(DegToRad((float)Loop));
    SinLockup[Loop] = sin(DegToRad((float)Loop));
  }

	// Initialize hardware for w buffering: (same as z buffering but
	// uses w instead of z)

  grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
	grDepthMask(FXTRUE);
	grDepthBufferFunction(GR_CMP_LESS);

  NearClippingPlane = 5.0;
  FarClippingPlane = 2500.0;

	TotNumLightsources = 0;

	HalfHorRes = HorRes/2;
	HalfVerRes = VerRes/2;


  // set Textures to NULL which indicates that no texture is loaded
  Textures = NULL;

	// get the startaddresses of the texture units:

	TexStartAddressTMU0 = grTexMinAddress(GR_TMU0);
	TexDownloadAddressTMU0 = TexStartAddressTMU0;
	TexMaxAddressTMU0 = grTexMaxAddress(GR_TMU0);

	if (GlobalDetailLevel != LOW_DETAIL)
	{
		TexStartAddressTMU1 = grTexMinAddress(GR_TMU1);
		TexDownloadAddressTMU1 = TexStartAddressTMU1;
		TexMaxAddressTMU1 = grTexMaxAddress(GR_TMU1);
	}

  // Set ambient light to 0,0,0
  AmbientR = 0.0;
  AmbientG = 0.0;
  AmbientB = 0.0;

  grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_CONSTANT,FXFALSE);

  grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,
      GR_BLEND_ZERO,GR_BLEND_ZERO);
  AlphaBlendMode = FXEFALSE;


  BackgroundColor.r = 0;
  BackgroundColor.g = 0;
  BackgroundColor.b = 0;

  CurrLandscape = NULL;

  DefaultCharacterset = NULL;

  HorInc = HorRes/640.0;
  VerInc = VerRes/480.0;

  ChromaKey = FXEFALSE;
  CombineTextureOnly = FXEFALSE;

  // Return success:
  return FXETRUE;
}


// This function should be called before your application terminates
// For the moment it just shut down glide but in the future new things
// could be destroyed here.
void fxeShutdown(void)
{
  // Shut down glide
  grGlideShutdown();
}

// Camera related functions:

void fxeCameraMove(float x, float y, float z)
{
	Camera.x = x;
	Camera.y = y;
	Camera.z = z;
}


void fxeCameraRotate(float ax, float ay, float az)
{
	Camera.ax = ax;
	Camera.ay = ay;
	Camera.az = az;
}

// Ambient light related functions:

void fxeAmbientlightLight(float red, float green, float blue)
{
  AmbientR = red;
  AmbientG = green;
  AmbientB = blue;
}

// Lightsource related functions:

// Allocates memory for a lightsource and inserts it into the
// linked list of lightsources
fxeLightsourceHandle fxeLightsourceAdd(float r, float g, float b,
  float x, float y, float z, float DistanceAffection)
{
	fxeLightsource *Lightsource;

  Lightsource = (fxeLightsource *)malloc(sizeof(fxeLightsource));
	Lightsource->r = r;
	Lightsource->g = g;
	Lightsource->b = b;
	Lightsource->x = x;
	Lightsource->y = y;
  Lightsource->z = z;
  Lightsource->lr = r/255.0;
  Lightsource->lg = g/255.0;
  Lightsource->lb = b/255.0;
  Lightsource->DistanceAffection = DistanceAffection;
	Lightsource->Next = NULL;

	// go to last lightsource in linked list of lightsources
	if (Lightsources == NULL)
  {
		// This is the first lightsource
		Lightsource->Prev = NULL;
	  Lightsources = Lightsource;
	}
	else
	{
		while (Lightsources->Next != NULL)
		{
			Lightsources = Lightsources->Next;
	  }
	  Lightsource->Prev = Lightsources;
    Lightsources->Next = Lightsource;
	  Lightsources = Lightsource;
	}
	return Lightsources;
}


void fxeLightsourceMove(fxeLightsourceHandle LightsourceHandle,
	float x, float y, float z)
{
	LightsourceHandle->x = x;
	LightsourceHandle->y = y;
	LightsourceHandle->z = z;
}


void fxeLightsourceLight(fxeLightsourceHandle LightsourceHandle,
	float r, float g, float b)
{
	LightsourceHandle->r = r;
	LightsourceHandle->g = g;
	LightsourceHandle->b = b;
  LightsourceHandle->lr = r/255.0;
  LightsourceHandle->lg = g/255.0;
  LightsourceHandle->lb = b/255.0;
}

// Note that if LightsourceHandle isn't a valid lightsource undefined
// results may occure such as a halting system crash!
void fxeLightsourceDelete(fxeLightsourceHandle LightsourceHandle)
{
	if (LightsourceHandle->Prev == NULL && LightsourceHandle->Next == NULL)
	{
		Lightsources = NULL;
	}
	else if (LightsourceHandle->Prev == NULL)
	{
		Lightsources = LightsourceHandle->Next;
		Lightsources->Prev = NULL;
  }
	else if (LightsourceHandle->Next == NULL)
	{
		Lightsources = LightsourceHandle->Prev;
		Lightsources->Next = NULL;
	}
	else
	{
		Lightsources = LightsourceHandle->Prev;
		Lightsources->Next = LightsourceHandle->Next;
		Lightsources = LightsourceHandle->Next;
		Lightsources->Prev = LightsourceHandle->Prev;
	}
	free(LightsourceHandle);
}

// Texture related functions:

// Loads a 3df texture file from disk, returns
// NULL if failure and handle to texture if
// successful.

fxeTextureHandle fxeTextureLoad(char *filename)
{
	Gu3dfInfo TexInfo;
  fxeTexture *Texture;



	// get the texture info:
  if (gu3dfGetInfo(filename,&TexInfo) == FXFALSE)
    return NULL;

  // allocate memory for the mipmap:
  TexInfo.data = (void*)malloc(TexInfo.mem_required);
  if (TexInfo.data == NULL)
    return NULL;

  // Load the mipmap from disk:
  gu3dfLoad(filename,&TexInfo);


	Texture = (fxeTexture*)malloc(sizeof(fxeTexture));
	if (Texture == NULL)
	{
    free(TexInfo.data);
		return NULL;
	}

	Texture->TexInfo.smallLod = TexInfo.header.small_lod;
	Texture->TexInfo.largeLod = TexInfo.header.large_lod;
	Texture->TexInfo.aspectRatio = TexInfo.header.aspect_ratio;
  Texture->TexInfo.data = TexInfo.data;
	Texture->TexInfo.format = TexInfo.header.format;

  if (GlobalDetailLevel == LOW_DETAIL)
	{
    Texture->FilteringMethod = BILINEAR_FILTERED;


		if ((TexDownloadAddressTMU0+grTexTextureMemRequired
			(GR_MIPMAPLEVELMASK_BOTH,&Texture->TexInfo)) <=
			TexMaxAddressTMU0)
		{
		grTexDownloadMipMap(GR_TMU0,TexDownloadAddressTMU0,
  	GR_MIPMAPLEVELMASK_BOTH,&Texture->TexInfo);

		Texture->VideoMemoryAddressTMU0 = TexDownloadAddressTMU0;

		TexDownloadAddressTMU0 = TexDownloadAddressTMU0+
			grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH,
			&Texture->TexInfo);
		}

		else
		{
			free(TexInfo.data);
			free(Texture);
			return NULL;
		}
	}

	else
	{
    Texture->FilteringMethod = TRILINEAR_FILTERED;

		if ((TexDownloadAddressTMU0+grTexTextureMemRequired
			(GR_MIPMAPLEVELMASK_EVEN,&Texture->TexInfo)) <=
			TexMaxAddressTMU0 && (TexDownloadAddressTMU1+
			grTexTextureMemRequired(GR_MIPMAPLEVELMASK_ODD,&
			Texture->TexInfo)) <= TexMaxAddressTMU1)
		{
		grTexDownloadMipMap(GR_TMU0,TexDownloadAddressTMU0,
  	GR_MIPMAPLEVELMASK_EVEN,&Texture->TexInfo);

  	grTexDownloadMipMap(GR_TMU1,TexDownloadAddressTMU1,
  	GR_MIPMAPLEVELMASK_ODD,&Texture->TexInfo);

		Texture->VideoMemoryAddressTMU0 = TexDownloadAddressTMU0;
		Texture->VideoMemoryAddressTMU1 = TexDownloadAddressTMU1;

		TexDownloadAddressTMU0 = TexDownloadAddressTMU0+
			grTexTextureMemRequired(GR_MIPMAPLEVELMASK_EVEN,
			&Texture->TexInfo);

		TexDownloadAddressTMU1 = TexDownloadAddressTMU1+
			grTexTextureMemRequired(GR_MIPMAPLEVELMASK_ODD,
			&Texture->TexInfo);
		}
		else
		{
			free(TexInfo.data);
			free(Texture);
			return NULL;
		}
	}

	if (Textures == NULL)
	{
		Texture->Prev = NULL;
		Texture->Next = NULL;
		Textures = Texture;
		return Textures;
	}

	else if (Textures->Next == NULL)
	{
		Textures->Next = Texture;
		Texture->Next = NULL;
		Texture->Prev = Textures;
		Textures = Texture;
		return Textures;
	}

	else if (Textures->Prev == NULL)
	{
    Texture->Next = Textures;
		Texture->Prev = NULL;
		Textures->Prev = Texture;
		Textures = Texture;
		return Textures;
	}

	else
	{
		Texture->Prev = Textures->Prev;
		Texture->Next = Textures;
		Textures->Prev = Texture;
		Textures = Texture->Prev;
		Textures->Next = Texture;
		Textures = Texture;
		return Textures;
	}
}

// Deletes a texture - TextureHandle must be a valid
// texturehandle or else strange things may happen!
fxeTextureDelete(fxeTextureHandle TextureHandle)
{
	if (TextureHandle->Next == NULL && TextureHandle->Prev
		== NULL)
		Textures = NULL;
	else if (TextureHandle->Next == NULL)
	{
		Textures = TextureHandle->Prev;
		Textures->Next = NULL;
	}
	else if (TextureHandle->Prev == NULL)
	{
		Textures = TextureHandle->Next;
		Textures->Prev = NULL;
	}
	else
	{
		Textures = TextureHandle->Prev;
		Textures->Next = TextureHandle->Next;
		Textures = TextureHandle->Next;
		Textures->Prev = TextureHandle->Prev;
	}

	free(TextureHandle->TexInfo.data);
	free(TextureHandle);
}




// Object related functions:

// Loads a object from disk and positions it at 0,0,0
// fxeObjectHandle fxeObjectLoad(char *filename);

// Adds an already created object to the worlds objects:
fxeObjectHandle fxeObjectAdd(fxeObject *Object)
{
	if (WorldObjects == NULL)
	{
		Object->Prev = NULL;
		Object->Next = NULL;
		WorldObjects = Object;
		return WorldObjects;
	}
	else
	{
		// Go to last object in world:
		if (WorldObjects->Next != NULL)
			WorldObjects = WorldObjects->Next;

		WorldObjects->Next = Object;
		Object->Prev = WorldObjects;
		Object->Next = NULL;
		WorldObjects = WorldObjects->Next;
		return WorldObjects;
	}
}

// Scales a object with SFactor.
void fxeObjectScale(fxeObjectHandle ObjectHandle, float SFactor)
{
  int Loop;
  fxeVertex *Vertex;

  Vertex = ObjectHandle->VertexList;

  for (Loop=0 ; Loop<ObjectHandle->NumVertices ; Loop++)
  {
    Vertex[Loop].x *= SFactor;
    Vertex[Loop].y *= SFactor;
    Vertex[Loop].z *= SFactor;
  }
}

// the angles is given in degress and must be in the range of
// 0..359
void fxeObjectRotate(fxeObjectHandle ObjectHandle,
  float xa, float ya, float za)
{
	int Loop;
	fxeVertex *Vertex;
  fxeMatrix XRot,YRot,ZRot,ConcatMatrix,ConcatMatrix2;

	if (xa == 0.0 && ya == 0.0 && za == 0.0) return;

	// we performs special logic depending on which angles we must rotate
	// around, when we only needs to rotate around one axis we dont need
	// than doing just that!

	// Check if only the x rotation matrix is required
	if (xa != 0.0 && ya == 0.0 && za == 0.0)
	{
		// Now we only need to multiply the x rotation matrix with every
		// vertex in the object:

    RotXMatrix(ConcatMatrix,360.0-xa);
	}

	// Check if only the y rotation matrix is required
	else if (xa == 0.0 && ya != 0.0 && za == 0.0)
	{
		// Now we only need to multiply the y rotation matrix with every
		// vertex in the object:

    RotYMatrix(ConcatMatrix,360.0-ya);
	}

	// Check if only the z rotation matrix is required
	else if (xa == 0.0 && ya == 0.0 && za != 0.0)
	{
		// Now we only need to multiply the z rotation matrix with every
		// vertex in the object:

    RotZMatrix(ConcatMatrix,360.0-za);
	}

	else if (xa != 0.0 && ya != 0.0 && za == 0.0)
	{
		// We must concatenate the x rotation matrix by the y rotation matrix:

    RotXMatrix(XRot,360.0-xa);
    RotYMatrix(YRot,360.0-ya);
		Matmult(XRot,YRot,ConcatMatrix);

	}

	else if (xa != 0.0 && ya == 0.0 && za != 0.0)
	{
		// We must concatenate the x rotation matrix by the z rotation matrix:

    RotXMatrix(XRot,360.0-xa);
    RotZMatrix(ZRot,360.0-za);
		Matmult(XRot,ZRot,ConcatMatrix);
	}

	else if (xa == 0.0 && ya != 0.0 && za != 0.0)
	{
		// We must concatenate the y rotation matrix by the z rotation matrix:

    RotYMatrix(YRot,360.0-ya);
    RotZMatrix(ZRot,360.0-za);
		Matmult(YRot,ZRot,ConcatMatrix);
	}

	else if (xa != 0.0 && ya != 0.0 && za != 0.0)
	{
		// We must concatenate the x rotation matrix by the y rotation matrix:

    RotXMatrix(XRot,360.0-xa);
    RotYMatrix(YRot,360.0-ya);
    RotZMatrix(ZRot,360.0-za);
		Matmult(XRot,YRot,ConcatMatrix2);
		Matmult(ConcatMatrix2,ZRot,ConcatMatrix);
	}

	Vertex = ObjectHandle->VertexList;

	if (Vertex != NULL)
  {

		for (Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
		{
      MatXVerASM(ConcatMatrix,&(Vertex[Loop]),&(Vertex[Loop]));
		}
	}
}


void fxeObjectMove(fxeObjectHandle ObjectHandle,
  float x, float y, float z)
{
	ObjectHandle->x = x;
	ObjectHandle->y = y;
	ObjectHandle->z = z;
}

// This function is not for use by a application, just internally.
FreeObjectMemory(fxeObjectHandle ObjectHandle)
{
	fxePolygon *Poly;
	int Loop;

	// go through all polygons and destroy respective polygons vertex
  // indices and texture coordinates if any
	Poly = ObjectHandle->PolygonList;
	if (Poly != NULL)
  {
		for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
		{
			free(Poly[Loop].VertexIndices);
      if (Poly[Loop].TexMode != NOT_TEXTURED)
        free(Poly[Loop].TextureCoordinates);
	  }

		// free polygon list
		free(ObjectHandle->PolygonList);
	}

  // Free all vertices:
	if (ObjectHandle->VertexList != NULL)
	{
		free(ObjectHandle->VertexList);
		free(ObjectHandle->VerticesWorld);
		free(ObjectHandle->VerticesCamera);
		free(ObjectHandle->VertexColors);
	}
}

// Note that if ObjectHandle isn't a valid Object handle undefined
// results may occure such as a halting system crash!
fxeObjectDelete(fxeObjectHandle ObjectHandle)
{
	if (ObjectHandle->Prev == NULL && ObjectHandle->Next == NULL)
	{
		WorldObjects = NULL;
	}
	else if (ObjectHandle->Prev == NULL)
	{
		WorldObjects = ObjectHandle->Next;
		WorldObjects->Prev = NULL;
  }
	else if (ObjectHandle->Next == NULL)
	{
		WorldObjects = ObjectHandle->Prev;
		WorldObjects->Next = NULL;
	}
	else
	{
		WorldObjects = ObjectHandle->Prev;
		WorldObjects->Next = ObjectHandle->Next;
		WorldObjects = ObjectHandle->Next;
		WorldObjects->Prev = ObjectHandle->Prev;
	}
	// destroy polygons and vertices associated with object
	FreeObjectMemory(ObjectHandle);
	free(ObjectHandle);
}

// Rendering associated functions:

// takes a objects local coordinates and transforms them
// to world coordinates:
void TransformToWorld(fxeObjectHandle ObjectHandle)
{
	fxeVertex *LocVer;
	fxeVertex *WorVer;
	int Loop;

	LocVer = ObjectHandle->VertexList;
	WorVer = ObjectHandle->VerticesWorld;


	if (LocVer != NULL)
	{
		for (Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
		{
			WorVer[Loop].x = LocVer[Loop].x + ObjectHandle->x;
			WorVer[Loop].y = LocVer[Loop].y + ObjectHandle->y;
			WorVer[Loop].z = LocVer[Loop].z + ObjectHandle->z;
			WorVer[Loop].w = LocVer[Loop].w;
    }
	}
}

void CalculateCameraTransformationMatrix(void)
{
  fxeMatrix Move,XRot,YRot,ZRot,Concat2,Concat3;

  // Now lets compute the transformation matrix:
  RotXMatrix(XRot,Camera.ax);
  RotYMatrix(YRot,Camera.ay);
  RotZMatrix(ZRot,Camera.az);
  MoveMatrix(Move,Camera.x,Camera.y,Camera.z);

  Matmult(Move,XRot,Concat3);
  Matmult(Concat3,YRot,Concat2);
  Matmult(Concat2,ZRot,CameraMatrix);
}


// This function clips invisible objects from the
// rendering pipeline - speeds up rendering a HUGE amount
void ClipObject(fxeObjectHandle ObjectHandle)
{
	float ObjectRadius;
	fxeVertex VerRelCamera;
	fxeVertex TempVertex;

	// When we shall clip objects we need to convert the objects
	// position in world to relative to camera

	// We need to invert the camera angles because when we
	// rotates the object to relative the camera position we must do
	// this in the inverse direction of the camera angles.

	ObjectHandle->Clipped = FXEFALSE;
  ObjectHandle->CompletelyVisible = FXETRUE;


  TempVertex.x = ObjectHandle->x;
  TempVertex.y = ObjectHandle->y;
  TempVertex.z = ObjectHandle->z;
  TempVertex.w = 1.0;

  MatXVer(CameraMatrix,&TempVertex,&VerRelCamera);


	// Rotate the position of the object in the inverse angle of
	// the camera:


	// now we have the position of the object in camera coordinates -
	// lets test the objects bounding box against the viewing volume:

  // Get the radius of the object:
  ObjectRadius = ObjectHandle->ObjectRadius;

  if ((VerRelCamera.z + ObjectRadius) < NearClippingPlane) 
    ObjectHandle->Clipped = FXETRUE;

  else if ((VerRelCamera.z - ObjectRadius) > FarClippingPlane)
		ObjectHandle->Clipped = FXETRUE;

  else if ((VerRelCamera.x + ObjectRadius) < (VerRelCamera.z*(-HFOWInc)))
  	ObjectHandle->Clipped = FXETRUE;

  else if ((VerRelCamera.x - ObjectRadius) > VerRelCamera.z*HFOWInc)
		ObjectHandle->Clipped = FXETRUE;

  else if ((VerRelCamera.y + ObjectRadius) < (VerRelCamera.z*(-VFOWInc)))
		ObjectHandle->Clipped = FXETRUE;

  else if ((VerRelCamera.y - ObjectRadius) > (VerRelCamera.z*VFOWInc))
		ObjectHandle->Clipped = FXETRUE;

}


void CrossProduct3d(fxeVector *Result, fxeVector *Vector1,
	fxeVector *Vector2)
{
	Result->x = (Vector1->y)*(Vector2->z)-(Vector1->z)*(Vector2->y);
	Result->y = -((Vector1->x)*(Vector2->z)-(Vector1->z)*(Vector2->x));
	Result->z = (Vector1->x)*(Vector2->y)-(Vector1->y)*(Vector2->x);
}

float DotProduct3d(fxeVector *Vector1, fxeVector *Vector2)
{
	return Vector1->x*Vector2->x + Vector1->y*Vector2->y +
		Vector1->z*Vector2->z;
}

float VectorLenght(fxeVector *Vector)
{
	return
    fast_fsqrt(Vector->x*Vector->x+Vector->y*Vector->y+Vector->z*Vector->z);
}


// Calculates normals for all polygons in object
void CalculateNormals(fxeObjectHandle ObjectHandle)
{
	int Loop;
	fxePolygon *Poly;
	fxeVector Vector1, Vector2;
	fxeVertexIndex *VIndex;
	fxeVertex *Vertex;
  fxeVertex *CurrVertex, *CurrVertex2, *CurrVertex3;

	Poly = ObjectHandle->PolygonList;
	Vertex = ObjectHandle->VerticesWorld;

  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
    if (Poly[Loop].Clipped == FXEFALSE)
    {
			VIndex = Poly[Loop].VertexIndices;
      CurrVertex = &(Vertex[VIndex[2]]);
      CurrVertex2 = &(Vertex[VIndex[1]]);
      CurrVertex3 = &(Vertex[VIndex[0]]);
      Vector1.x = CurrVertex->x-CurrVertex3->x;
      Vector1.y = CurrVertex->y-CurrVertex3->y;
      Vector1.z = CurrVertex->z-CurrVertex3->z;
      Vector2.x = CurrVertex2->x-CurrVertex3->x;
      Vector2.y = CurrVertex2->y-CurrVertex3->y;
      Vector2.z = CurrVertex2->z-CurrVertex3->z;

      CrossProduct3dASM(&(Poly[Loop].Normal),&Vector1,&Vector2);
    }
	}
}


void RemoveBackfaces(fxeObjectHandle ObjectHandle)
{
	int Loop;
	fxePolygon *Poly;
	fxeVector LineOfSight;
	fxeVertexIndex *VIndex;
	fxeVertex *Vertex;

	Poly = ObjectHandle->PolygonList;
  Vertex = ObjectHandle->VerticesWorld;

  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
    // Check if the polygon is doublesided:

    Poly[Loop].Culled = FXEFALSE;

    if (Poly[Loop].Clipped == FXEFALSE)
    {

      if (Poly[Loop].DoubleSided == FXEFALSE)
      {

        // Calculate the line of sight vector, the vector
        // going from one point of the polygon to
        // the camera position:

        VIndex = Poly[Loop].VertexIndices;

        LineOfSight.x = Camera.x-Vertex[VIndex[0]].x;
        LineOfSight.y = Camera.y-Vertex[VIndex[0]].y;
        LineOfSight.z = Camera.z-Vertex[VIndex[0]].z;

        // Make Dot product with the normal of the polygon
        // and the light of sight vector, if the sign
        // of the result is positive then the polygon
        // is visible, if not it's invisible:

        if (DotProduct3d(&LineOfSight,&(Poly[Loop].Normal))
          < 0.0)
        {
          Poly[Loop].Culled = FXETRUE;
          ObjectHandle->CompletelyVisible = FXEFALSE;
        }
      }
		}
  }
}


fxeVector *GetPolygonNormal(fxeVertex *Vertex);


void ShadeObject(fxeObjectHandle ObjectHandle)
{
	// This is propably one of the most complex functions in the 3d engine
 	// since it must take into considereration an unlimited amount of
	// lightsources and perform a lot of calculations.

	fxePolygon *Poly;
	fxeVertex *Vertex;
	fxeVertex *CamVertex;
  fxeVertex *VertexList;
	fxeVector LightVector;
	fxeColor *VertexColor;
	fxeVector AddedNormal;
	fxeVertexIndex *VIndex;
  float NormalVectorLenght;
  float LightVectorLenght;
  float Dp3d;
	float ReflectionIntensity;
  fxeColor *ConstantColor;
  fxeColor *FlatShadedColor;
  float TotalShadeR, TotalShadeG, TotalShadeB;
  fxeVertex *CurrVertex;
  fxeVector *CurrVector;
  fxeVector *Normal;
  float DistanceAffection;
  float SpecularIntensity;

  int Loop, Loop2, Finished;

	Finished = FXEFALSE;

	Poly = ObjectHandle->PolygonList;
	Vertex = ObjectHandle->VerticesWorld;
  VertexList = ObjectHandle->VertexList;
	CamVertex = ObjectHandle->VerticesCamera;
	VertexColor = ObjectHandle->VertexColors;
	Poly = ObjectHandle->PolygonList;


  if (ObjectHandle->ShadingType == FLAT_SHADING)
  {
    // Go through all polygons:
    for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
    {
      VIndex = Poly[Loop].VertexIndices;
      ConstantColor = &(Poly[Loop].ConstantColor);
      FlatShadedColor = &(Poly[Loop].FlatShadedColor);
      FlatShadedColor->r = 0.0;
      FlatShadedColor->g = 0.0;
      FlatShadedColor->b = 0.0;
      Normal = &(Poly[Loop].Normal);
      NormalVectorLenght = VectorLenght(Normal);

      if (Lightsources != NULL)
			{
				// go to first lightsource in world:
				while (Lightsources->Prev != NULL)
					Lightsources = Lightsources->Prev;

				Finished = FXEFALSE;
				// Loop through all lightsources
				while (Finished == FXEFALSE)
				{
          // Calculate reflection level of this Polygon
					// with the current lightsource

					// Calculate lightsource vector:
          LightVector.x = Lightsources->x - Vertex[VIndex[0]].x;
          LightVector.y = Lightsources->y - Vertex[VIndex[0]].y;
          LightVector.z = Lightsources->z - Vertex[VIndex[0]].z;

					// Calculate lenght of lightsource vector:
          // add lightvectorLenght with 0.001 to avoid division by zero
          LightVectorLenght = VectorLenght(&LightVector)+0.001;

					// Calculate the dot product with the lightsource vector
          // and the normal vector for this polygon

          Dp3d = LightVector.x*Normal->x+LightVector.y*Normal->y+
            LightVector.z*Normal->z;

					// Divide the addednormal by addednormal lenght


					// calculate the level of light that this vertex reflects
					// from the current lightsource: (scale it to 0..1)

          //if (Poly[Loop].DoubleSided == FXEFALSE)
          //{

            ReflectionIntensity = ((Dp3d /
              (LightVectorLenght * NormalVectorLenght)) + 1.0) *0.5;

          //}

          /*else
          {
            ReflectionIntensity = (Dp3d /
              (LightVectorLenght * NormalVectorLenght));
            if (ReflectionIntensity < 0.0)
              ReflectionIntensity = -ReflectionIntensity;
          }*/

          // we adds the reflection level the polygon gives
					// from the current lightsource to the total reflection level
				  // which is stored in the vertex structure of the object
					// As FxEngine uses RGB lightning the RGB intensity components
					// from the lightsource is used. In the formula also the
					// lenght of the light vector is used so that lightsources
          // near polygon does lit the polygon more and lightsources far
          // away from the polygon lits the polygon less.

          DistanceAffection =
            (Lightsources->DistanceAffection*LightVectorLenght);

          TotalShadeR =
            (ConstantColor->r*(ReflectionIntensity*Lightsources->lr))-
            DistanceAffection;

          TotalShadeG =
            (ConstantColor->g*(ReflectionIntensity*Lightsources->lg))-
            DistanceAffection;

          TotalShadeB =
            (ConstantColor->b*(ReflectionIntensity*Lightsources->lb))-
            DistanceAffection;

          if (TotalShadeR < 0.0)
            TotalShadeR = 0.0;

          if (TotalShadeG < 0.0)
            TotalShadeG = 0.0;

          if (TotalShadeB < 0.0)
            TotalShadeB = 0.0;

          FlatShadedColor->r += TotalShadeR;
          FlatShadedColor->g += TotalShadeG;
          FlatShadedColor->b += TotalShadeB;


					if (Lightsources->Next == NULL)
						Finished = FXETRUE;
					else
						Lightsources = Lightsources->Next;
				}
          
       

			}

        if (FlatShadedColor->r < 0.0)
          FlatShadedColor->r = 0.0;
        if (FlatShadedColor->g < 0.0)
          FlatShadedColor->g = 0.0;
        if (FlatShadedColor->b < 0.0)
          FlatShadedColor->b = 0.0;

        FlatShadedColor->r += AmbientR;
        FlatShadedColor->g += AmbientG;
        FlatShadedColor->b += AmbientB;

				// Make sure light of the Vertex is always smaller or
				// equal to 255.0
        if (FlatShadedColor->r > 255.0)
          FlatShadedColor->r = 255.0;

        if (FlatShadedColor->g > 255.0)
          FlatShadedColor->g = 255.0;

        if (FlatShadedColor->b > 255.0)
          FlatShadedColor->b = 255.0;

    }
  }

  else if (ObjectHandle->ShadingType == GOURAUD_SHADING)
  {
		// Loop through the vertex list of the object -
		// check which polygons who shares the vertex, calculate
		// normals for polygon1,polygon2,..,polygonN who shares the vertex
		// and sum these normals together. A dot product between this
		// normalsum and every light source is performed to calculate
		// the RGB intensity of every vertex in the object.
		for (Loop=0; Loop<ObjectHandle->NumVertices; Loop++)
		{
      AddedNormal.x = 0.0;
      AddedNormal.y = 0.0;
      AddedNormal.z = 0.0;
      VertexColor[Loop].r = 0.0;
      VertexColor[Loop].g = 0.0;
      VertexColor[Loop].b = 0.0;

      CurrVertex = &(VertexList[Loop]);

      for (Loop2=0; Loop2<CurrVertex->NumPolys; Loop2++)
      {
        CurrVector =
          &(Poly[CurrVertex->PolygonIndices[Loop2]].Normal);

        AddedNormal.x += CurrVector->x;
        AddedNormal.y += CurrVector->y;
        AddedNormal.z += CurrVector->z;
      }



//      AddedNormal2.x = AddedNormal.x/NormalVectorLenght;
//      AddedNormal2.y = AddedNormal.y/NormalVectorLenght;

/*
    for (Loop2=0; Loop2<ObjectHandle->NumPolys; Loop2++)
    {
      if (Poly[Loop2].TexMode == ENVIROMENT_MAPPED &&
          Poly[Loop2].Clipped == FXEFALSE)
				{
					VIndex = Poly[Loop2].VertexIndices;

					for (Loop3=0; Loop3<Poly[Loop2].NumVertices; Loop3++)
					{
						if (VIndex[Loop3] == Loop)
						{
							TextureHandle = Poly[Loop2].TextureHandle;
							switch(TextureHandle->TexInfo.aspectRatio)
							{
								case GR_ASPECT_8x1:
									TextureCenterX = 128.0;
									TextureCenterY = 16.0;
									break;
								case GR_ASPECT_4x1:
									TextureCenterX = 128.0;
									TextureCenterY = 32.0;
									break;
								case GR_ASPECT_2x1:
									TextureCenterX = 128.0;
									TextureCenterY = 64.0;
									break;
								case GR_ASPECT_1x1:
									TextureCenterX = 128.0;
									TextureCenterY = 128.0;
									break;
								case GR_ASPECT_1x2:
									TextureCenterX = 64.0;
									TextureCenterY = 128.0;
									break;
								case GR_ASPECT_1x4:
									TextureCenterX = 32.0;
									TextureCenterY = 128.0;
									break;
								case GR_ASPECT_1x8:
									TextureCenterX = 16.0;
									TextureCenterY = 128.0;
									break;
							}

							Poly[Loop2].TextureCoordinates[Loop3].s =
								TextureCenterX+AddedNormal2.x*TextureCenterX;
							Poly[Loop2].TextureCoordinates[Loop3].t =
								TextureCenterY+AddedNormal2.y*TextureCenterY;
						}
					}
				}
			}
*/

      // Do the dot product between this total normal vector and
      // all lightsources in the 3d world


			// Now we should have a total normal vector for this vertex
      NormalVectorLenght = VectorLenght(&AddedNormal);

      if (Lightsources != NULL)
      {
        // go to first lightsource in world:
        while (Lightsources->Prev != NULL)
          Lightsources = Lightsources->Prev;

        Finished = FXEFALSE;
        // Loop through all lightsources
        while (Finished == FXEFALSE)
        {
          // Calculate reflection level of this vertex
          // with the current lightsource

          // Calculate lightsource vector:
          LightVector.x = Lightsources->x - Vertex[Loop].x;
          LightVector.y = Lightsources->y - Vertex[Loop].y;
          LightVector.z = Lightsources->z - Vertex[Loop].z;

          // Calculate lenght of lightsource vector:
          // add lightvectorLenght with 0.001 to avoid division by zero

          LightVectorLenght = VectorLenght(&LightVector)+0.001;

          // Calculate the dot product with the lightsource vector
          // and the added normal vector for this vertex

          Dp3d = LightVector.x*AddedNormal.x+LightVector.y*AddedNormal.y+
            LightVector.z*AddedNormal.z;


          // calculate the level of light that this vertex reflects
          // from the current lightsource: (scale it to 0..1)

//          if (Poly[Loop].DoubleSided == FXEFALSE)
//          {

            ReflectionIntensity = ((Dp3d /
              (LightVectorLenght * NormalVectorLenght)) + 1.0) *0.5;
//          }

/*          else
          {
            ReflectionIntensity = (Dp3d /
              (LightVectorLenght * NormalVectorLenght));
            if (ReflectionIntensity < 0.0)
              ReflectionIntensity = -ReflectionIntensity;
          } */

          // we adds the reflection level the vertex gives
          // from the current lightsource to the total reflection level
          // which is stored in the vertex structure of the object
          // As FxEngine uses RGB lightning the RGB intensity components
          // from the lightsource is used. In the formula also the
          // lenght of the light vector is used so that lightsources
          // near vertex does lit the vertex more and lightsources far
          // away from the verted lits the vertex less.

          DistanceAffection = (LightVectorLenght*Lightsources->DistanceAffection);

          TotalShadeR = (Lightsources->r*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeG = (Lightsources->g*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeB = (Lightsources->b*ReflectionIntensity)
            -DistanceAffection;

          if (TotalShadeR < 0.0)
            TotalShadeR = 0.0;

          if (TotalShadeG < 0.0)
            TotalShadeG = 0.0;

          if (TotalShadeB < 0.0)
            TotalShadeB = 0.0;


          VertexColor[Loop].r += TotalShadeR;
          VertexColor[Loop].g += TotalShadeG;
          VertexColor[Loop].b += TotalShadeB;

          if (Lightsources->Next == NULL)
            Finished = FXETRUE;
          else
            Lightsources = Lightsources->Next;
        }


      }
        VertexColor[Loop].r += AmbientR;
        VertexColor[Loop].g += AmbientG;
        VertexColor[Loop].b += AmbientB;

        // Make sure light of the Vertex is always smaller or
        // equal to 255.0 and bigger or equal to 0.0
        if (VertexColor[Loop].r > 255.0)
          VertexColor[Loop].r = 255.0;

        if (VertexColor[Loop].g > 255.0)
          VertexColor[Loop].g = 255.0;

        if (VertexColor[Loop].b > 255.0)
          VertexColor[Loop].b = 255.0;

    }
  }

  else if (ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
  {
		// Loop through the vertex list of the object -
		// check which polygons who shares the vertex, calculate
		// normals for polygon1,polygon2,..,polygonN who shares the vertex
		// and sum these normals together. A dot product between this
		// normalsum and every light source is performed to calculate
		// the RGB intensity of every vertex in the object.
		for (Loop=0; Loop<ObjectHandle->NumVertices; Loop++)
		{
      AddedNormal.x = 0.0;
      AddedNormal.y = 0.0;
      AddedNormal.z = 0.0;
      VertexColor[Loop].r = 0.0;
      VertexColor[Loop].g = 0.0;
      VertexColor[Loop].b = 0.0;

      CurrVertex = &(VertexList[Loop]);

      for (Loop2=0; Loop2<CurrVertex->NumPolys; Loop2++)
      {
        CurrVector =
          &(Poly[CurrVertex->PolygonIndices[Loop2]].Normal);

        AddedNormal.x += CurrVector->x;
        AddedNormal.y += CurrVector->y;
        AddedNormal.z += CurrVector->z;
      }


			// Now we should have a total normal vector for this vertex
      NormalVectorLenght = VectorLenght(&AddedNormal);

      // Do the dot product between this total normal vector and
      // all lightsources in the 3d world

      if (Lightsources != NULL)
      {
        // go to first lightsource in world:
        while (Lightsources->Prev != NULL)
          Lightsources = Lightsources->Prev;

        Finished = FXEFALSE;
        // Loop through all lightsources
        while (Finished == FXEFALSE)
        {
          // Calculate reflection level of this vertex
          // with the current lightsource

          // Calculate lightsource vector:
          LightVector.x = Lightsources->x - Vertex[Loop].x;
          LightVector.y = Lightsources->y - Vertex[Loop].y;
          LightVector.z = Lightsources->z - Vertex[Loop].z;

          // Calculate lenght of lightsource vector:
          // add lightvectorLenght with 0.001 to avoid division by zero

          LightVectorLenght = VectorLenght(&LightVector)+0.001;

          // Calculate the dot product with the lightsource vector
          // and the added normal vector for this vertex

          Dp3d = LightVector.x*AddedNormal.x+LightVector.y*AddedNormal.y+
            LightVector.z*AddedNormal.z;




          // calculate the level of light that this vertex reflects
          // from the current lightsource: (scale it to 0..1)

//          if (Poly[Loop].DoubleSided == FXEFALSE)
//          {
              ReflectionIntensity =
                (Dp3d/(LightVectorLenght*NormalVectorLenght) + 1.0) * 0.5;
              SpecularIntensity =
                255.0*pow(ReflectionIntensity,ObjectHandle->SpecConstant);
              ReflectionIntensity += SpecularIntensity;

              if (ReflectionIntensity > 1.0)
                ReflectionIntensity = 1.0;

//          }

/*          else
          {
            ReflectionIntensity = (Dp3d /
              (LightVectorLenght * NormalVectorLenght));

            if (ReflectionIntensity < 0.0)
              ReflectionIntensity = -ReflectionIntensity;
          } */

          // we adds the reflection level the vertex gives
          // from the current lightsource to the total reflection level
          // which is stored in the vertex structure of the object
          // As FxEngine uses RGB lightning the RGB intensity components
          // from the lightsource is used. In the formula also the
          // lenght of the light vector is used so that lightsources
          // near vertex does lit the vertex more and lightsources far
          // away from the verted lits the vertex less.

          DistanceAffection = (LightVectorLenght*Lightsources->DistanceAffection);

          TotalShadeR = (Lightsources->r*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeG = (Lightsources->g*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeB = (Lightsources->b*ReflectionIntensity)
            -DistanceAffection;

          if (TotalShadeR < 0.0)
            TotalShadeR = 0.0;

          if (TotalShadeG < 0.0)
            TotalShadeG = 0.0;

          if (TotalShadeB < 0.0)
            TotalShadeB = 0.0;


          VertexColor[Loop].r += TotalShadeR;
          VertexColor[Loop].g += TotalShadeG;
          VertexColor[Loop].b += TotalShadeB;

          if (Lightsources->Next == NULL)
            Finished = FXETRUE;
          else
            Lightsources = Lightsources->Next;
        }

        VertexColor[Loop].r += AmbientR;
        VertexColor[Loop].g += AmbientG;
        VertexColor[Loop].b += AmbientB;

        // Make sure light of the Vertex is always smaller or
        // equal to 255.0 and bigger or equal to 0.0
        if (VertexColor[Loop].r > 255.0)
          VertexColor[Loop].r = 255.0;

        if (VertexColor[Loop].g > 255.0)
          VertexColor[Loop].g = 255.0;

        if (VertexColor[Loop].b > 255.0)
          VertexColor[Loop].b = 255.0;

      }
    }
  }

}




void TransformToCameraMatrix(fxeObjectHandle ObjectHandle)
{
	fxeVertex *VerticesWorld;
	fxeVertex *VerticesCamera;
	int Loop;

	VerticesWorld = ObjectHandle->VerticesWorld;
	VerticesCamera = ObjectHandle->VerticesCamera;

	for(Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
	{
    MatXVerASM(CameraMatrix,&(VerticesWorld[Loop]),&(VerticesCamera[Loop]));
	}
}

// Clips polygons in camera space:
void ClipPolygons(fxeObjectHandle ObjectHandle)
{
  int Loop, Loop2;
  fxePolygon *PolygonList;
  fxeVertex *VerticesCamera;
  int NumberClippedX1;
  int NumberClippedY1;
  int NumberClippedZ1;
  int NumberClippedX2;
  int NumberClippedY2;
  int NumberClippedZ2;
  int ClippedNearPlane;
  int NumVertices;
  fxeVertexIndex *VertexIndices;
  fxeVertex *CurrVertex;
  float x, y, z, Radius;

  PolygonList = ObjectHandle->PolygonList;
  VerticesCamera = ObjectHandle->VerticesCamera;


  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
    PolygonList[Loop].Clipped = FXEFALSE;
    PolygonList[Loop].PartlyClipped = FXEFALSE;
    PolygonList[Loop].PartlyClippedZ = FXEFALSE;

    VertexIndices = PolygonList[Loop].VertexIndices;

    x = VerticesCamera[VertexIndices[0]].x;
    y = VerticesCamera[VertexIndices[0]].y;
    z = VerticesCamera[VertexIndices[0]].z;
    Radius = PolygonList[Loop].PolygonRadius;

    if (((z-Radius) < NearClippingPlane) || ((z+Radius) > FarClippingPlane) ||
      ((x-Radius) < (-z)) || ((x+Radius) > z) ||
      ((y-Radius) < (z*(-VFOWInc))) || ((y+Radius) > (z*(VFOWInc))))
    {


    NumVertices = PolygonList[Loop].NumVertices;

    NumberClippedX1 = 0;
    NumberClippedY1 = 0;
    NumberClippedZ1 = 0;
    NumberClippedX2 = 0;
    NumberClippedY2 = 0;
    NumberClippedZ2 = 0;

    ClippedNearPlane = FXEFALSE;


    for (Loop2=0; Loop2<NumVertices; Loop2++)
    {
      CurrVertex = &(VerticesCamera[VertexIndices[Loop2]]);

  
      if ((CurrVertex->z) < NearClippingPlane)
        NumberClippedZ1++;

      if ((CurrVertex->z) > FarClippingPlane)
        NumberClippedZ2++;

      if ((CurrVertex->x) < (-CurrVertex->z))
        NumberClippedX1++;

      if ((CurrVertex->x) > (CurrVertex->z))
        NumberClippedX2++;

      if ((CurrVertex->y) < (CurrVertex->z*(-VFOWInc)))
        NumberClippedY1++;

      if ((CurrVertex->y) > (CurrVertex->z*VFOWInc))
        NumberClippedY2++;

    }

    if (NumberClippedX1 == NumVertices || NumberClippedX2 == NumVertices ||
      NumberClippedY1 == NumVertices || NumberClippedY2 == NumVertices ||
      NumberClippedZ1 == NumVertices || NumberClippedZ2 == NumVertices)
    {
        ObjectHandle->CompletelyVisible = FXEFALSE;
        PolygonList[Loop].Clipped = FXETRUE;
    }

    else if (NumberClippedX1 > 0 || NumberClippedX2 > 0 ||
      NumberClippedY1 > 0 || NumberClippedY2 > 0 ||
      NumberClippedZ1 > 0 || NumberClippedZ2 > 0)
    {
      // The polygon is only partly clipped
      PolygonList[Loop].PartlyClipped = FXETRUE;
      ObjectHandle->CompletelyVisible = FXEFALSE;



      if (NumberClippedZ1 > 0) 
        PolygonList[Loop].PartlyClippedZ = FXETRUE;
    }
    }
  }
}

void ClipPolygonsToNearPlane(fxeObjectHandle ObjectHandle)
{
  int NumPolygons;
  fxeColor *ClippedVertexColors;
  fxeColor *VertexColors;
  fxeColor *CurrVertexColor;
  int Loop, Loop2;
  fxePolygon *PolygonList;
  fxeVertex *VerticesCamera;
  fxeVertexIndex *VertexIndices;
  fxeVertex *CurrVertex;
  fxeTextureCoordinate *TextureCoordinates;
  fxeTextureCoordinate *CurrTexCoor;
  float MultValue;
  int NumVertices;
  int DestVertex;
  int Nextvert;
  int Curin, Nextin;
  int CurZ, NextZ;
  fxeVertex *ClippedVertices;
  fxeTextureCoordinate *ClippedTextureCoordinates;


  PolygonList = ObjectHandle->PolygonList;
  NumPolygons = ObjectHandle->NumPolys;
  VerticesCamera = ObjectHandle->VerticesCamera;
  VertexColors = ObjectHandle->VertexColors;

  for (Loop=0; Loop<NumPolygons ; Loop++)
  {
    if (PolygonList[Loop].PartlyClippedZ == FXETRUE &&
      PolygonList[Loop].Culled == FXEFALSE)
    {
        // Go through all edges of the polygon

        NumVertices = PolygonList[Loop].NumVertices;

        PolygonList[Loop].NumAddedVertices = 0;

        DestVertex = 0;

        ClippedVertices = PolygonList[Loop].ClippedVertices;

        ClippedVertexColors = PolygonList[Loop].ClippedVertexColors;

        TextureCoordinates = PolygonList[Loop].TextureCoordinates;

        ClippedTextureCoordinates =
          PolygonList[Loop].ClippedTextureCoordinates;

        VertexIndices = PolygonList[Loop].VertexIndices;

        CurZ = VerticesCamera[VertexIndices[0]].z;
        Curin = (CurZ >= NearClippingPlane);


        for (Loop2=0; Loop2<NumVertices; Loop2++)
        {
          Nextvert = (Loop2 + 1) % NumVertices;

          if (Curin)
          {
            CurrVertex = &(VerticesCamera[VertexIndices[Loop2]]);
            CurrVertexColor = &(VertexColors[VertexIndices[Loop2]]);

            ClippedVertices[DestVertex].x = CurrVertex->x;
            ClippedVertices[DestVertex].y = CurrVertex->y;
            ClippedVertices[DestVertex].z = CurrVertex->z;
            
            if (PolygonList[Loop].TexMode != NOT_TEXTURED)
            {
              CurrTexCoor = &(TextureCoordinates[Loop2]);
              ClippedTextureCoordinates[DestVertex].s = CurrTexCoor->s;
              ClippedTextureCoordinates[DestVertex].t = CurrTexCoor->t;
            }

            ClippedVertexColors[DestVertex].r = CurrVertexColor->r;
            ClippedVertexColors[DestVertex].g = CurrVertexColor->g;
            ClippedVertexColors[DestVertex].b = CurrVertexColor->b;

            PolygonList[Loop].NumAddedVertices++;

            DestVertex++;
          }

          NextZ = VerticesCamera[VertexIndices[Nextvert]].z;
          Nextin = (NextZ >= NearClippingPlane);

          if (Curin != Nextin)
          {
            MultValue = (NearClippingPlane - CurZ) /
              (NextZ - CurZ);

            CurrVertex = &(VerticesCamera[VertexIndices[Loop2]]);
            CurrVertexColor = &(VertexColors[VertexIndices[Loop2]]);

            ClippedVertices[DestVertex].x = CurrVertex->x +
              ((VerticesCamera[VertexIndices[Nextvert]].x - CurrVertex->x)
              * MultValue);

            ClippedVertices[DestVertex].y = CurrVertex->y +
              ((VerticesCamera[VertexIndices[Nextvert]].y - CurrVertex->y)
              * MultValue);
            
            ClippedVertices[DestVertex].z = CurrVertex->z +
              ((VerticesCamera[VertexIndices[Nextvert]].z - CurrVertex->z)
              * MultValue);

            if (PolygonList[Loop].TexMode != NOT_TEXTURED)
            {
              CurrTexCoor = &(TextureCoordinates[Loop2]);

              ClippedTextureCoordinates[DestVertex].s = CurrTexCoor->s +
                ((TextureCoordinates[Nextvert].s - CurrTexCoor->s)
                * MultValue);

              ClippedTextureCoordinates[DestVertex].t = CurrTexCoor->t +
                ((TextureCoordinates[Nextvert].t - CurrTexCoor->t)
                * MultValue);
            }

            ClippedVertexColors[DestVertex].r = CurrVertexColor->r;


            ClippedVertexColors[DestVertex].g = CurrVertexColor->g;

            ClippedVertexColors[DestVertex].b = CurrVertexColor->b;

            PolygonList[Loop].NumAddedVertices++;

            DestVertex++;
          }

          CurZ = NextZ;
          Curin = Nextin;
        }
      }
    }
}

int SortPolygonsCallback(fxePolygonPtr *ppPolygon1,
  fxePolygonPtr *ppPolygon2)
{
  int Loop;
  fxePolygonPtr pPolygon1 = *ppPolygon1;
  fxePolygonPtr pPolygon2 = *ppPolygon2;
  fxeVertex *VerticesCamera;
  float AverageZ1 = 0;
  float AverageZ2 = 0;

  VerticesCamera = CurrentObject->VerticesCamera;

  // Calculate average z for polygon:


  for (Loop=0; Loop<pPolygon1->NumVertices; Loop++)
  {
    AverageZ1 += VerticesCamera[pPolygon1->VertexIndices[Loop]].z;
  }

  for (Loop=0; Loop<pPolygon2->NumVertices; Loop++)
  {
    AverageZ2 += VerticesCamera[pPolygon2->VertexIndices[Loop]].z;
  }

  AverageZ1 = AverageZ1 / (float)(pPolygon1->NumVertices);
  AverageZ2 = AverageZ2 / (float)(pPolygon2->NumVertices);

  if (AverageZ1 > AverageZ2)
    return -1;
  else if (AverageZ1 < AverageZ2)
    return 1;
  else
    return 0;
}

void ProjectAndDrawObject(fxeObjectHandle ObjectHandle)
{
	// Note!
  // DrawVertex is a array of 200 vertices, it polygon has
  // more than 200 vertices this function will bail out!

  GrVertex DrawVertex[MAX_VERTICES_PER_POLY];
	int Loop, Loop2;
	fxePolygon *Poly;
  fxeVertex *CamVertex, *ProjectedVertices;
	fxeColor *VertexColor;
	fxeVertexIndex *VIndex;
	fxeTextureHandle TexHandle;
  int NumVertices;
  int ClippedNearPlane;
  fxeVertex *ClippedVertices;
  fxeTextureCoordinate *ClippedTexCoors;
  fxeColor *ClippedVertexColors;
  fxeColor *FlatShadedColor;
  float ooz, oozView;
  fxeVertex *CurrVertex;
  fxeColor *CurrVertexColor;
  float sow, tow;
  fxePolygon *CurrPoly;
  GrVertex *CurrDrawVertex;
  GrTmuVertex *tmuvtx0, *tmuvtx1;
  float px,py;
  fxeVertex *CurrProjVertex;


	Poly = ObjectHandle->PolygonList;
	CamVertex = ObjectHandle->VerticesCamera;
	VertexColor = ObjectHandle->VertexColors;
  ProjectedVertices = ObjectHandle->ProjectedVertices;

        if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
          ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
        {
          if (ObjectHandle->ObjectType == OBJECT_COMPLETELY_TEXTURED)
          {
            // Setup color combine unit for gouraud shading
            // and texture mapping:
            grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
              GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
              GR_COMBINE_OTHER_TEXTURE,FXFALSE);
          }
          else if (ObjectHandle->ObjectType == OBJECT_NOT_TEXTURED)
          {
            // Setup color combine unit for gouraud shading
            grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
              GR_COMBINE_FACTOR_NONE,GR_COMBINE_LOCAL_ITERATED,
              GR_COMBINE_OTHER_NONE,FXFALSE);
          }
        }

        else if (ObjectHandle->ShadingType == FLAT_SHADING)
        {

          if (ObjectHandle->ObjectType == OBJECT_COMPLETELY_TEXTURED)
          {
            // Setup color combine unit for flat shading
            // and texture mapping:
            grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
              GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_CONSTANT,
              GR_COMBINE_OTHER_TEXTURE,FXFALSE);
          }
          else if (ObjectHandle->ObjectType == OBJECT_NOT_TEXTURED)
          {
            // Setup color combine unit for flat shading
            grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
              GR_COMBINE_FACTOR_NONE,GR_COMBINE_LOCAL_CONSTANT,
              GR_COMBINE_OTHER_NONE,FXFALSE);
          }
        }

 

  if (ObjectHandle->CompletelyVisible == FXETRUE)
  {
    // Go through all vertices in polygon and transform them
    // to screen space:

    for (Loop=0; Loop<ObjectHandle->NumVertices; Loop++)
    {
          CurrVertex = &(CamVertex[Loop]);
          CurrProjVertex = &(ProjectedVertices[Loop]);

          ooz = 1.0 / CurrVertex->z;
          oozView = ooz * ViewingDistance;
          px = HalfHorRes+oozView*CurrVertex->x;
          px += SnapValue;
          px -= SnapValue;
          py = HalfVerRes+oozView*CurrVertex->y;
          py += SnapValue;
          py -= SnapValue;

          CurrProjVertex->x = px;
          CurrProjVertex->y = py;
          CurrProjVertex->z = ooz;
    }
  }


  // Setup alpha blend unit to no alpha blending:

  if (AlphaBlendMode == FXETRUE && ObjectHandle->Transparent == FXEFALSE)
  {
    grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,
      GR_BLEND_ZERO,GR_BLEND_ZERO);
    AlphaBlendMode = FXEFALSE;
  }

	for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
	{
    
    // only project and draw polygon if visible
    CurrPoly = &(Poly[Loop]);

    if (CurrPoly->Transparent == FXEFALSE)
    {
      // We draws the transparent last as they need the rgb values
      // from the other polygons rendered in the linear frame buffer

    if (CurrPoly->Clipped == FXEFALSE && CurrPoly->Culled == FXEFALSE)
		{
      VIndex = CurrPoly->VertexIndices;

      if (CurrPoly->PartlyClippedZ == FXETRUE)
      {
        NumVertices = CurrPoly->NumAddedVertices;
        ClippedNearPlane = FXETRUE;
        ClippedVertexColors = CurrPoly->ClippedVertexColors;
        ClippedVertices = CurrPoly->ClippedVertices;
        ClippedTexCoors = CurrPoly->ClippedTextureCoordinates;
      }
      else
      {
        NumVertices = CurrPoly->NumVertices;
        ClippedNearPlane = FXEFALSE;
      }

      for (Loop2=0; Loop2<NumVertices; Loop2++)
			{

				// Now it's time to project the polygon on the screen with
				// perspective projection. Note - must snap the vertices
				// so that 3dfx hardware doesn't mess up!

        if (ClippedNearPlane == FXETRUE)
        {
          CurrVertex = &(ClippedVertices[Loop2]);
          ooz = 1.0 / CurrVertex->z;
          oozView = ooz * ViewingDistance;

          DrawVertex[Loop2].x = Snap(HalfHorRes+CurrVertex->x*oozView);
          DrawVertex[Loop2].y = Snap(HalfVerRes+CurrVertex->y*oozView);
          DrawVertex[Loop2].oow = ooz;

          if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
            ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
          {
            DrawVertex[Loop2].r = ClippedVertexColors[Loop2].r;
            DrawVertex[Loop2].g = ClippedVertexColors[Loop2].g;
            DrawVertex[Loop2].b = ClippedVertexColors[Loop2].b;
          }

          // If this polygon shall be texture mapped then
          // copy texture coordinates.
          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            if (Poly[Loop].TextureHandle->FilteringMethod ==
              BILINEAR_FILTERED)
            {
              sow = ClippedTexCoors[Loop2].s*ooz;
              tow = ClippedTexCoors[Loop2].t*ooz;
              // we just need to copy texture coordinates for
              // TMU0:
              DrawVertex[Loop2].tmuvtx[0].sow = sow;
              DrawVertex[Loop2].tmuvtx[0].tow = tow;


            }

            else if (CurrPoly->TextureHandle->FilteringMethod
              == TRILINEAR_FILTERED)
            {
              sow = ClippedTexCoors[Loop2].s*ooz;
              tow = ClippedTexCoors[Loop2].t*ooz;

              // we must copy texture coordinates for TMU0 and
              // TMU1:
              DrawVertex[Loop2].tmuvtx[0].sow = sow;
              DrawVertex[Loop2].tmuvtx[0].tow = tow;
              DrawVertex[Loop2].tmuvtx[1].sow = sow;
              DrawVertex[Loop2].tmuvtx[1].tow = tow;
            }
          }
        }

        else
        {
          if (ObjectHandle->CompletelyVisible == FXEFALSE)
          {
            CurrVertex = &(CamVertex[VIndex[Loop2]]);
            CurrDrawVertex = &(DrawVertex[Loop2]);

            ooz = 1.0 / CurrVertex->z;
            oozView = ooz * ViewingDistance;
            px = HalfHorRes+oozView*CurrVertex->x;
            px += SnapValue;
            px -= SnapValue;
            py = HalfVerRes+oozView*CurrVertex->y;
            py += SnapValue;
            py -= SnapValue;

            CurrDrawVertex->x = px;

            CurrDrawVertex->y = py;

            CurrDrawVertex->oow = ooz;
          }

          else
          {
            CurrVertex = &(ProjectedVertices[VIndex[Loop2]]);
            CurrDrawVertex = &(DrawVertex[Loop2]);
            
            CurrDrawVertex->x = CurrVertex->x;
            CurrDrawVertex->y = CurrVertex->y;
            CurrDrawVertex->oow = CurrVertex->z;
            ooz = CurrVertex->z;
          }

          if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
            ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
          {
            CurrVertexColor = &(VertexColor[VIndex[Loop2]]);
            CurrDrawVertex->r = CurrVertexColor->r;
            CurrDrawVertex->g = CurrVertexColor->g;
            CurrDrawVertex->b = CurrVertexColor->b;
          }

          // If this polygon shall be texture mapped then
          // copy texture coordinates.
          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            if (CurrPoly->TextureHandle->FilteringMethod ==
              BILINEAR_FILTERED)
            {
              sow = CurrPoly->TextureCoordinates[Loop2].s*ooz;
              tow = CurrPoly->TextureCoordinates[Loop2].t*ooz;
              tmuvtx0 = &(CurrDrawVertex->tmuvtx[0]);

              // we just need to copy texture coordinates for
              // TMU0:
              tmuvtx0->sow = sow;
              tmuvtx0->tow = tow;
            }

            else if (CurrPoly->TextureHandle->FilteringMethod
              == TRILINEAR_FILTERED)
            {
              sow = CurrPoly->TextureCoordinates[Loop2].s*ooz;
              tow = CurrPoly->TextureCoordinates[Loop2].t*ooz;
              tmuvtx0 = &(CurrDrawVertex->tmuvtx[0]);
              tmuvtx1 = &(CurrDrawVertex->tmuvtx[1]);

              // we must copy texture coordinates for TMU0 and
              // TMU1:
              tmuvtx0->sow = sow;
              tmuvtx0->tow = tow;
              tmuvtx1->sow = sow;
              tmuvtx1->tow = tow;

            }
          }
        }
			}


        if (CurrPoly->TexMode != NOT_TEXTURED)
        {

          TexHandle = CurrPoly->TextureHandle;

          // Setup texture combine unit(s) and write
          // texture coordinates to DrawVertex

          // check which texture filtering method which
          // is used:
          if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
          {
            // Setup texture unit 0's texture source
            grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
              GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
          }

          else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
          {
            // Combine trilinear texture mapping with
            // bilinear filtering to produce trilinear
            // filtering

            // Setup texture unit 0's texture source
            grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
              GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

            // Setup texture unit 1's texture source
            grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
              GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
          }
        }

        if (ObjectHandle->ShadingType == FLAT_SHADING)
        {
          FlatShadedColor = &(CurrPoly->FlatShadedColor);
          grConstantColorValue( ((int)(FlatShadedColor->r)<<16) |
            ((int)(FlatShadedColor->g)<<8) | (int)(FlatShadedColor->b));
        }

        // so finally send polygon to 3dfx hardware:
        // Draw clipped if needed:

        if (CurrPoly->PartlyClipped == FXEFALSE)
          grDrawPolygonVertexList(NumVertices,DrawVertex);
        else
          guDrawPolygonVertexListWithClip(NumVertices,DrawVertex);
		}
    }
  }


  // Setup alpha blending unit for alpha blending of polygons
  // with background scene.

  if (ObjectHandle->Transparent == FXETRUE)
  {
    if (AlphaBlendMode == FXEFALSE && ObjectHandle->Transparent == FXETRUE)
    {
      grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
        GR_BLEND_ONE,GR_BLEND_ZERO);
      AlphaBlendMode = FXETRUE;
    }


    // We must sort the polygons in Z since transparent polygons needs
    // this:


    for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
    {
      PolygonPointers[Loop] = &(ObjectHandle->PolygonList[Loop]);
    }

    CurrentObject = ObjectHandle;

    qsort(PolygonPointers,ObjectHandle->NumPolys,
      sizeof(fxePolygonPtr),SortPolygonsCallback);



  for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
	{
    
    // only project and draw polygon if visible
    CurrPoly = PolygonPointers[Loop];

    // Now we project and draws all transparent polygons:

    if (CurrPoly->Transparent == FXETRUE)
    {
      // We draws the transparent last as they need the rgb values
      // from the other polygons rendered in the linear frame buffer

    if (CurrPoly->Clipped == FXEFALSE && CurrPoly->Culled == FXEFALSE)
		{
      VIndex = CurrPoly->VertexIndices;

      if (CurrPoly->PartlyClippedZ == FXETRUE)
      {
        NumVertices = CurrPoly->NumAddedVertices;
        ClippedNearPlane = FXETRUE;
        ClippedVertexColors = CurrPoly->ClippedVertexColors;
        ClippedVertices = CurrPoly->ClippedVertices;
        ClippedTexCoors = CurrPoly->ClippedTextureCoordinates;
      }
      else
      {
        NumVertices = CurrPoly->NumVertices;
        ClippedNearPlane = FXEFALSE;
      }

      for (Loop2=0; Loop2<NumVertices; Loop2++)
			{

				// Now it's time to project the polygon on the screen with
				// perspective projection. Note - must snap the vertices
				// so that 3dfx hardware doesn't mess up!

        if (ClippedNearPlane == FXETRUE)
        {
          CurrVertex = &(ClippedVertices[Loop2]);
          ooz = 1.0 / CurrVertex->z;
          oozView = ooz * ViewingDistance;

          DrawVertex[Loop2].x = Snap(HalfHorRes+CurrVertex->x*oozView);
          DrawVertex[Loop2].y = Snap(HalfVerRes+CurrVertex->y*oozView);
          DrawVertex[Loop2].oow = ooz;

          if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
            ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
          {
            DrawVertex[Loop2].r = ClippedVertexColors[Loop2].r;
            DrawVertex[Loop2].g = ClippedVertexColors[Loop2].g;
            DrawVertex[Loop2].b = ClippedVertexColors[Loop2].b;
          }

          // If this polygon shall be texture mapped then
          // copy texture coordinates.
          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            if (Poly[Loop].TextureHandle->FilteringMethod ==
              BILINEAR_FILTERED)
            {
              sow = ClippedTexCoors[Loop2].s*ooz;
              tow = ClippedTexCoors[Loop2].t*ooz;
              // we just need to copy texture coordinates for
              // TMU0:
              DrawVertex[Loop2].tmuvtx[0].sow = sow;
              DrawVertex[Loop2].tmuvtx[0].tow = tow;


            }

            else if (CurrPoly->TextureHandle->FilteringMethod
              == TRILINEAR_FILTERED)
            {
              sow = ClippedTexCoors[Loop2].s*ooz;
              tow = ClippedTexCoors[Loop2].t*ooz;

              // we must copy texture coordinates for TMU0 and
              // TMU1:
              DrawVertex[Loop2].tmuvtx[0].sow = sow;
              DrawVertex[Loop2].tmuvtx[0].tow = tow;
              DrawVertex[Loop2].tmuvtx[1].sow = sow;
              DrawVertex[Loop2].tmuvtx[1].tow = tow;
            }
          }
        }

        else
        {
          if (ObjectHandle->CompletelyVisible == FXEFALSE)
          {
            CurrVertex = &(CamVertex[VIndex[Loop2]]);
            CurrDrawVertex = &(DrawVertex[Loop2]);

            ooz = 1.0 / CurrVertex->z;
            oozView = ooz * ViewingDistance;
            px = HalfHorRes+oozView*CurrVertex->x;
            px += SnapValue;
            px -= SnapValue;
            py = HalfVerRes+oozView*CurrVertex->y;
            py += SnapValue;
            py -= SnapValue;

            CurrDrawVertex->x = px;

            CurrDrawVertex->y = py;

            CurrDrawVertex->oow = ooz;
          }

          else
          {
            CurrVertex = &(ProjectedVertices[VIndex[Loop2]]);
            CurrDrawVertex = &(DrawVertex[Loop2]);
            
            CurrDrawVertex->x = CurrVertex->x;
            CurrDrawVertex->y = CurrVertex->y;
            CurrDrawVertex->oow = CurrVertex->z;
            ooz = CurrVertex->z;
          }

          if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
            ObjectHandle->ShadingType == GOURAUD_SHADING_WSH)
          {
            CurrVertexColor = &(VertexColor[VIndex[Loop2]]);
            CurrDrawVertex->r = CurrVertexColor->r;
            CurrDrawVertex->g = CurrVertexColor->g;
            CurrDrawVertex->b = CurrVertexColor->b;
          }

          // If this polygon shall be texture mapped then
          // copy texture coordinates.
          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            if (CurrPoly->TextureHandle->FilteringMethod ==
              BILINEAR_FILTERED)
            {
              sow = CurrPoly->TextureCoordinates[Loop2].s*ooz;
              tow = CurrPoly->TextureCoordinates[Loop2].t*ooz;
              tmuvtx0 = &(CurrDrawVertex->tmuvtx[0]);

              // we just need to copy texture coordinates for
              // TMU0:
              tmuvtx0->sow = sow;
              tmuvtx0->tow = tow;
            }

            else if (CurrPoly->TextureHandle->FilteringMethod
              == TRILINEAR_FILTERED)
            {
              sow = CurrPoly->TextureCoordinates[Loop2].s*ooz;
              tow = CurrPoly->TextureCoordinates[Loop2].t*ooz;
              tmuvtx0 = &(CurrDrawVertex->tmuvtx[0]);
              tmuvtx1 = &(CurrDrawVertex->tmuvtx[1]);

              // we must copy texture coordinates for TMU0 and
              // TMU1:
              tmuvtx0->sow = sow;
              tmuvtx0->tow = tow;
              tmuvtx1->sow = sow;
              tmuvtx1->tow = tow;

            }
          }
        }
			}


        if (CurrPoly->TexMode != NOT_TEXTURED)
        {

          TexHandle = CurrPoly->TextureHandle;

          // Setup texture combine unit(s) and write
          // texture coordinates to DrawVertex

          // check which texture filtering method which
          // is used:
          if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
          {
            // Setup texture unit 0's texture source
            grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
              GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
          }

          else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
          {
            // Combine trilinear texture mapping with
            // bilinear filtering to produce trilinear
            // filtering

            // Setup texture unit 0's texture source
            grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
              GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

            // Setup texture unit 1's texture source
            grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
              GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
          }
        }

        if (ObjectHandle->ShadingType == FLAT_SHADING)
        {
          FlatShadedColor = &(CurrPoly->FlatShadedColor);
          grConstantColorValue((((int)CurrPoly->AlphaLevel)<<24) |
            ((int)(FlatShadedColor->r)<<16) |
            ((int)(FlatShadedColor->g)<<8) | (int)(FlatShadedColor->b));
        }
        else
        {
          grConstantColorValue((((int)CurrPoly->AlphaLevel)<<24));
        }


        // so finally send polygon to 3dfx hardware:
        // Draw clipped if needed:

        if (CurrPoly->PartlyClipped == FXEFALSE)
          grDrawPolygonVertexList(NumVertices,DrawVertex);
        else
          guDrawPolygonVertexListWithClip(NumVertices,DrawVertex);
		}
    }
  }
  }
}

int Sort_Objects_Callback_Function(fxeObjectPtr *ppObject1,
  fxeObjectPtr *ppObject2)
{
  // this function is called by qsort and sorts all polygons using simple
  // z value.
  fxeVertex TempVertex;
  fxeVertex VerRelCamera;
  fxeVertex TempVertex2;
  fxeVertex VerRelCamera2;
  fxeObjectPtr pObject1 = *ppObject1;
  fxeObjectPtr pObject2 = *ppObject2;

  TempVertex.x = pObject1->x;
  TempVertex.y = pObject1->y;
  TempVertex.z = pObject1->z;
  TempVertex.w = 1.0;
  TempVertex2.x = pObject2->x;
  TempVertex2.y = pObject2->y;
  TempVertex2.z = pObject2->z;
  TempVertex2.w = 1.0;

  MatXVer(CameraMatrix,&TempVertex,&VerRelCamera);
  MatXVer(CameraMatrix,&TempVertex2,&VerRelCamera2);


  if (VerRelCamera.z > VerRelCamera2.z)
    return -1;
  else if (VerRelCamera.z < VerRelCamera2.z)
    return 1;
  else
    return 0;
}

void fxeBackgroundSetColor(fxeColor *pColor)
{
  BackgroundColor.r = pColor->r;
  BackgroundColor.g = pColor->g;
  BackgroundColor.b = pColor->b;
}

// This is the big one. It transforms all objects to world space and
// then to view space. It calculates light reflection of all objects.
// It performs object clipping and polygon clipping and backface removal.
// Perspective projection also.
// 2d clipping it performed by glide but fxeWorldRender() calls
// different glide rendering functions depending on if the polygon
// needs to be clipped or not.
// Glide sorts the polygons with W buffering.
// Also it checks if textures needs to be used and sets up texture units
// if needed.

void fxeWorldRender(int DisplayWorld)
{
	int Finished;
  int NumObjects;
  int Loop;

  FrameDisplayed = FXEFALSE;
  
  grBufferClear( ((int)(BackgroundColor.r)<<16)+
    ((int)(BackgroundColor.g)<<8)+((int)(BackgroundColor.b)),
    0,GR_WDEPTHVALUE_FARTHEST);

  grChromakeyMode(GR_CHROMAKEY_DISABLE);
  ChromaKey = FXEFALSE;
  CombineTextureOnly = FXEFALSE;

  if (CurrLandscape != NULL)
    fxeLandscapeRender(CurrLandscape);

	if (WorldObjects != NULL)
	{
		// go to first object in world:
		while (WorldObjects->Prev != NULL)
	 		WorldObjects = WorldObjects->Prev;

		Finished = FXEFALSE;

    CalculateCameraTransformationMatrix();

		// go through all objects and process them and render them

    NumObjects = 0;

  
    while(Finished == FXEFALSE)
		{

      ClipObject(WorldObjects);

      if (WorldObjects->Clipped == FXEFALSE)
      {
        WorldObjectPointers[NumObjects] = WorldObjects;
        NumObjects++;
      }

			if (WorldObjects->Next != NULL)
				WorldObjects = WorldObjects->Next;
			else
				Finished = FXETRUE;
    }

    // Sort the object list:

    qsort(WorldObjectPointers,NumObjects,sizeof(fxeObjectPtr),
      Sort_Objects_Callback_Function);

    for (Loop=0; Loop<NumObjects; Loop++)
    {
      TransformToWorld(WorldObjectPointers[Loop]);

      TransformToCameraMatrix(WorldObjectPointers[Loop]);
      ClipPolygons(WorldObjectPointers[Loop]);
      CalculateNormals(WorldObjectPointers[Loop]);
      ShadeObject(WorldObjectPointers[Loop]);
      RemoveBackfaces(WorldObjectPointers[Loop]);
      ClipPolygonsToNearPlane(WorldObjectPointers[Loop]);
      ProjectAndDrawObject(WorldObjectPointers[Loop]);
    }

	}

  if (DisplayWorld == FXETRUE)
  {
    fxeWorldDisplay();
    FrameDisplayed = FXETRUE;
  }

  FrameCounter++;
}

void fxeBackgroundClear(void)
{
  grBufferClear( ((int)(BackgroundColor.r)<<16)+
    ((int)(BackgroundColor.g)<<8)+((int)(BackgroundColor.b)),
    0,GR_WDEPTHVALUE_FARTHEST);
}

void fxeWorldDisplay(void)
{
  if (FrameDisplayed == FXEFALSE)
  {
    // Sync to 60FPS
    grBufferSwap(1);
    FrameDisplayed = FXETRUE;
  }
}


// Here comes some matrix math functions by Hugues Landry which also is
// going to be used in a 3d engine called Sp@cE_iNvAdr.
// Big thanks to you Hugues for this!
// Original Matrix functions by Hugues Landry,
// modification for FxEngine by Andreas Ingo

//////////////////////////////////////////////
// Transformation Module Function Definiton //
//////////////////////////////////////////////

// Initialization Function Definition

void InitMatrix(fxeMatrix Matrix)
{
	Matrix[0][0]=1;	Matrix[1][0]=0;	Matrix[2][0]=0; Matrix[3][0]=0;
	Matrix[0][1]=0;	Matrix[1][1]=1; Matrix[2][1]=0; Matrix[3][1]=0;
	Matrix[0][2]=0;	Matrix[1][2]=0;	Matrix[2][2]=1; Matrix[3][2]=0;
	Matrix[0][3]=0;	Matrix[1][3]=0; Matrix[2][3]=0; Matrix[3][3]=1;
}


// Matrix Transformation Function Definitions

void MoveMatrix(fxeMatrix Matrix, float x, float y, float z)
{
	InitMatrix(Matrix);
  Matrix[3][0] = -x;
  Matrix[3][1] = -y;
  Matrix[3][2] = -z;
}

// Note! The rotation matrix functions requires that CosLockup and
// SinLockup have been setup. Angle is the angle in degrees.
// These functions builds a inverted rotation matrices. To rotate in
// non inverted direction specify Angle argument as
// 360.0 - (Real Angle)


void RotXMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

	Matrix[1][1] =  CosLockup[(int)(Angle)];
  Matrix[1][2] = -SinLockup[(int)(Angle)];
  Matrix[2][1] = -(-SinLockup[(int)(Angle)]);
	Matrix[2][2] =  CosLockup[(int)(Angle)];
}


void RotYMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

	Matrix[0][0] =  CosLockup[(int)(Angle)];
  Matrix[0][2] = -(-SinLockup[(int)(Angle)]);
  Matrix[2][0] = -(SinLockup[(int)(Angle)]);
	Matrix[2][2] =  CosLockup[(int)(Angle)];
}

void RotZMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

	Matrix[0][0] =  CosLockup[(int)(Angle)];
  Matrix[0][1] = -(SinLockup[(int)(Angle)]);
  Matrix[1][0] = -(-SinLockup[(int)(Angle)]);
	Matrix[1][1] =  CosLockup[(int)(Angle)];
}

void ScaleMatrix(fxeMatrix Matrix, float SFactor)
{
	InitMatrix(Matrix);

	Matrix[0][0] = SFactor;
	Matrix[1][1] = SFactor;
	Matrix[2][2] = SFactor;
}

void ProjectMatrix(fxeMatrix Matrix)
{
	InitMatrix(Matrix);

	Matrix[3][2] = 1;
	Matrix[3][3] = 0;
}

// Matrix multiplication Function Definitions

void Matmult(fxeMatrix MatrixA, fxeMatrix MatrixB, fxeMatrix Result)
{
  int i,j,k;
  float Sum;

  for (i=0 ; i<4 ; i++)
  {
    for (j=0 ; j<4 ; j++)
    {
      Sum = 0;

      for (k=0 ; k < 4 ; k++)
        Sum += MatrixA[i][k] * MatrixB[k][j];

      Result[i][j] = Sum;
    }
  }
}


void MatXVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result)
{

    Result->x = Vertex->x * Matrix[0][0];
    Result->x += Vertex->y * Matrix[1][0];
    Result->x += Vertex->z * Matrix[2][0];
    Result->x += Matrix[3][0];

    Result->y = Vertex->x * Matrix[0][1];
    Result->y += Vertex->y * Matrix[1][1];
    Result->y += Vertex->z * Matrix[2][1];
    Result->y += Matrix[3][1];

    Result->z = Vertex->x * Matrix[0][2];
    Result->z += Vertex->y * Matrix[1][2];
    Result->z += Vertex->z * Matrix[2][2];
    Result->z += Matrix[3][2];

    Result->w = 1.0;
}



